#include "Analytic.h"

double adjustRadii(double vol, double param1, double param2, double param3) {
	return vol * param1;
}

//******************************************************************************
void Analytic::calculate(Molecule &m, NeckParams &neckParams, double A, double B, double alpha,
 double beta, double gamma, double delta, double omega ) {

	Atoms::iterator i, j;

	//Radius of the solvent, aka water
        double Rw = neckParams.WaterRadius(); //this probe radius is used for Necks only 
	double R_sphere = 10.0;

        // Neck Parameters
        int p = 4;
        int q = 4;
        Node * n ;



	for (i = m.begin(); i != m.end(); i++) {

		//Get the radius of atom "i"
                double R_i = i -> OrigRadius();

                // Set the region for chunks and NoNecks 
                set<int>::iterator itr;
		for (itr = i->noneckBegin(); itr != i->noneckEnd() ; itr++ ){
                     m.SetAtomRegion(*itr , 2  );
		}
		for (itr = i->chunkBegin(); itr != i->chunkEnd(); itr++ ){
                     m.SetAtomRegion(*itr , 1  );
		}

		// Contribution of the chunk considered as a collection of spheres
		// Compute Pairwise Contribution and necks for atom i
                double approx_sum_ch = 0.0;
		double sum_pw = 0.0; 
		double sum_neck = 0.0;
		double VolumeMeasure=0 ;
				
		for (j = m.begin(); j != m.end(); j++ ) {

                        // Get Radius of atom "j" and distance i and j
                        double R_j = j -> OrigRadius();
                        double r_ij = i -> distance( *j);

                        // Atoms in the VDW region . 
                        if ( j->Region == 0 ) {

                                //Scale down radii to account for overlaps. Same scale
                                //factor for all; Equivalent to fs in Amber.
                                double b_j = delta*R_j;
                                //Evaluate Grycuk function (sphere contribution to alpha i) 
                                sum_pw  = sum_pw + Functions::Grycuk(R_i, b_j, r_ij);

                                // Necks
				//n  = neckParams.lookup(R_i, R_j);
                                //double Aij = n->Aij;
                                //double Bij = n->Bij;


                                //If the atoms are too far, there is no neck, i.e. add 0
                                if (r_ij <= R_i + R_j + 2*Rw) {
                                	// Get Necks parameters
                                	n  = neckParams.lookup(R_i, R_j);
                                	double Aij = n->Aij;
                                	double Bij = n->Bij;
					if (r_ij >= Bij) sum_neck += Aij*pow(r_ij-Bij, p)*pow(R_i+R_j+2*Rw - r_ij, q);
                                }


                        }

                        // CHunk region
                        else if ( j->Region == 1 ) {
                                //Scale down radii to account for overlaps. Same scale
                                //factor for all; Equivalent to fs in Amber.
                                approx_sum_ch = approx_sum_ch + Functions::Grycuk(R_i, R_j, r_ij);

                        }

                        // Atoms that overlap with the Chunks dont have necks
			else if ( j->Region == 2 ) {
                                //Scale down radii to account for overlaps. Same scale
                                //factor for all; Equivalent to fs in Amber.
                                double b_j = delta*R_j;
                                //Evaluate Grycuk function (sphere contribution to alpha i) 
                                sum_pw  = sum_pw + Grycuk(R_i, b_j, r_ij);
			}


			//comput a measure of Volume  to see how deep an atom is
			if  ( (r_ij <= R_sphere) &&  (i->Number() != j->Number())) {

				VolumeMeasure = VolumeMeasure + Functions::SmoothFactor(R_sphere,  r_ij)*R_j*R_j*R_j ;
			}

		}		

		double sum_ch = (i->ChunkR6Packing())*approx_sum_ch;
		double rho3 = R_i*R_i*R_i;
		double rho3tilda = 1/rho3 - sum_ch;

		sum_neck = 3./(4.*M_PI)*sum_neck;

		double I = sum_pw + omega*sum_neck;

		double resultNoTanh = Functions::CubeRoot(1/(R_i*R_i*R_i) - I - sum_ch);
		double c_i = 1 - (1/(A*A*A*rho3tilda));
		double phi_i = I * (1/rho3tilda);
		double alpha = 1/c_i;

		double VolFactor = VolumeMeasure / (R_sphere * R_sphere * R_sphere );
		double phi2_i = rho3 * I * VolFactor; 
		//double phi2_i = phi_i * VolFactor; 

		double result = Functions::CubeRoot(rho3tilda - c_i * rho3tilda * tanh(phi_i*alpha - beta*phi2_i*phi2_i + gamma*phi2_i*phi2_i*phi2_i))+B;
		i->MolecularR6Sum(result);

// 		cout << i->Number() << "\t" << result  << endl;
//		cout << i->Number() << "\t" << sum_pw  << "\t" << sum_neck << endl;

		for (itr = i->noneckBegin(); itr != i->noneckEnd() ; itr++ ){
                     m.SetAtomRegion(*itr , 0  );
		}

	}
}


//////////////////////////////////////////////////////////////
// Grycuk function
// Modified B.A
double Analytic::Grycuk(double b_i, double b_j, double r_ij) {

 	/*b_i swallowed b_j, don't count anything.*/
        if (b_i > r_ij + b_j) { return 0.0 ; }

        /*b_j swallowed by b_i, this should be counted. Right*/
        /*now, however, we don't.*/
        if (b_j > r_ij + b_i) {
        	cerr << "Atom swalloed in Grycuk: b_i=" << b_i
        	 << " r_ij=" << r_ij << " b_j=" << b_j << endl;
		return 0.0;
        }
        /*Atoms are far enough apart, there isn't any overlap*/
        else if (r_ij > b_i + b_j) {
        	double term = b_j / (r_ij*r_ij - b_j*b_j);
                return term*term*term;
        }
        /*Atoms are overlapping.*/
        else {
        	double v0 = 16 * r_ij;
                double v1 = r_ij + 3*b_j;
                double v2 = r_ij + b_j;
                double v3 = 3*(b_j*b_j - b_i*b_i - (r_ij-b_i)*(r_ij-b_i))+2*r_ij*b_i;
                double v4 = b_i*b_i*b_i*b_i;
                return (1./v0)*(v1/(v2*v2*v2)+v3/v4);
        }
}

//////////////////////////////////////////////////////////////
// compute the number of necks
// Created by B.A
// Quick code to count number of necks for each atoms
void Analytic::NumberOfNecks(Molecule &m, NeckParams &neckParams, double A, double B, double alpha,
 double beta, double gamma, double delta, double omega, Molecule Group[], int NGroup ) {

	Atoms::iterator i, j;
	for (i = m.begin(); i != m.end(); i++) {
		Atom a_i = *i;
        	double Ri = a_i.OrigRadius();

		int NNecks = 0;
        	for (int k = 0 ; k < NGroup ; k++) {
			for (j = Group[k].begin(); j != Group[k].end(); j++) {
                		/*Already accounted for atoms in chunk. Skip these.*/
                        	if (i->isInChunk(j->Number())) { continue;}

                        	Atom a_j = *j;
                        	double rij = a_i.distance(a_j);
                        	double Rj = a_j.Radius();
					
				double Rw = neckParams.WaterRadius();//this is used for in different Rw 
			
				if (rij <= Ri + Rj + 2*Rw)  {
                                	NNecks = NNecks + 1;
                        	}
		
                	}

        	}
		cout << i->Number()  << "\t" << NNecks << endl;
	}
}




//////////////////////////////////////////////////////////////
// Quick code to compute the number of probable Hydrogen bonds per Atom
// Created by Boris Aguilar 
//  
void Analytic::NumberOfHB(Molecule &m, NeckParams &neckParams, double A, double B, double alpha,
 double beta, double gamma, double delta, double omega, Molecule Group[], int NGroup ) {

        Atoms::iterator i, j;
        for (i = m.begin(); i != m.end(); i++) {
                Atom a_i = *i;
                double Ri = a_i.OrigRadius();

		cout << i -> Number() << " | ";

               	int NNecks = 0;
               	for (int k = 0 ; k < NGroup ; k++) {
               		for (j = Group[k].begin(); j != Group[k].end(); j++) {
                                /*Already accounted for atoms in chunk. Skip these.*/
                                if (i->isInChunk(j->Number())) { continue;}

                                Atom a_j = *j;
                                double Rj = a_j.Radius();		
                                double rij = a_i.distance(a_j);
                                double Rw = neckParams.WaterRadius();//this is used for in different Rw 
                                if (rij <= Ri + Rj + 2*Rw)  {

					if (Ri <= 1.3) { // This will compute necks only for hydrogen atoms
						set<int>::iterator D = a_i.neighborsBegin();
						Atom Donor = m.getAtom(*D);
						double R_d = Donor.Radius();
						
						if ((R_d == 1.500) && ( (Rj == 1.550) || (Rj == 1.500) )) {
							NNecks = NNecks + 1;
							cout << a_j.Number() << "  ";}
						else if ((R_d == 1.550) && ( (Rj == 1.550) || (Rj == 1.500) )) { 
							NNecks = NNecks + 1;
							cout << a_j.Number() << "  ";}
						else if ((R_d == 1.700) && ( Rj == 1.500)) { 
							NNecks = NNecks + 1;
							cout << a_j.Number() << "  ";}
					
					}
					else if ( (Ri == 1.500) && (Rj <=1.3) )   {
						set<int>::iterator D = a_j.neighborsBegin();
						Atom Donor = m.getAtom(*D);
						double R_d = Donor.Radius();
						
						if (R_d == 1.500) {
							NNecks = NNecks +1;
							cout << a_j.Number() << "  ";}
						else if (R_d == 1.5500) {
							NNecks = NNecks + 1;
							cout << a_j.Number() << "  ";}

					}
					else if ( (Ri == 1.550) && (Rj <=1.3) )   {
						set<int>::iterator D = a_j.neighborsBegin();
						Atom Donor = m.getAtom(*D);
						double R_d = Donor.Radius();
						
						if (R_d == 1.500) {
							NNecks = NNecks +1;
							cout << a_j.Number() << "  ";}
						else if (R_d == 1.5500) {
							NNecks = NNecks + 1;
							cout << a_j.Number() << "  ";}

						

					}
                                }
				
				

                        }

                }
	
		cout  << " | " << NNecks << endl;
        }
	
}

//////////////////////////////////////////////////////////////
// Compute the number of times the complicated parts of the grycuck functions 
// are used within the chunks.
// Boris Aguilar
void Analytic::GrycukAnalysis(Molecule &m, NeckParams &neckParams, double A, double B, double alpha,
 double beta, double gamma, double delta, double omega ) {

 Atoms::iterator i, j;
 for (i = m.begin(); i != m.end(); i++) {
	Atom a_i = *i;
        double R_i = a_i.OrigRadius();
	
	cout << "ATOM " << i->Number() << endl;
        for (j = m.begin(); j != m.end() ; j++ ) {
		if ((i->isInChunk(j->Number())) && (i->Number() != j->Number())  ) {
			if (! i->isInNeighbors(j->Number())) {
				Atom a_j = *j ;
				double R_j = a_j.OrigRadius();
				double r_ij = a_i.distance(a_j);
				if (r_ij < R_j + R_i ) {
					cout << j -> Number() << endl ;
				}
			}
		}
	}
 }  
}
