/*****************************************************************************
//Grycuk.h
//Created by Richard Shadrach
//2008-7-29
//
//Modified by Boris Aguilar 08/12/09
//
//
//Description:
//Calculates effective born radii.  It does so by computing the effective
//radii for the chunk through msms, then pairwise descreening with necks.
//
//Definitions:
//Chunk: See Atom.h
//alpha: Parameter for tanh trick
//beta:  Parameter for tanh trick
//gamma: Parameter for tanh trick
//delta: Weight of pairwise descreening.  Reduces radii to account for
//         complex overlaps which occur far away from atom i.
//omega: Weight of the neck.  This is multiplied by the value for the neck
//       to account for neck-neck and neck-atom overlaps.
*****************************************************************************/


#ifndef ANALYTIC_H
#define ANALYTIC_H

#include <vector>

using namespace std;

#include "Molecule.h"
#include "NeckParams.h"

class Analytic {
public:
	static void calculate(Molecule &m, NeckParams &neckParams, double A, double B=0, double alpha=1,
	 double beta=0., double gamma=0., double delta=1., double omega=1.);
	
	static void NumberOfNecks(Molecule &m, NeckParams &neckParams, double A, double B=0, double alpha=1,
	 double beta=0., double gamma=0., double delta=1., double omega=1., Molecule Group[] = NULL, int NGroup = 0);

	static void NumberOfHB(Molecule &m, NeckParams &neckParams, double A, double B=0, double alpha=1,
	 double beta=0., double gamma=0., double delta=1., double omega=1., Molecule Group[] = NULL, int NGroup = 0);

	static void GrycukAnalysis(Molecule &m, NeckParams &neckParams, double A, double B=0, double alpha=1,
	 double beta=0., double gamma=0., double delta=1., double omega=1.);

private:
	static double Necks(Molecule &m, Atom &a_i, NeckParams &neckParams, Molecule Group[], int NGroup);
        static double Grycuk(double b_i, double b_j, double rij);
};

#endif

