#include "Atom.h"	

/**************************Constructors**************************************/
Atom::Atom(double paramX, double paramY, double paramZ, double paramRadius,
 int paramNumber, double paramCharge, string paramType) {
	x = paramX;	y = paramY;	z = paramZ;
	radius = origRadius = paramRadius;
	number = paramNumber;
	charge = paramCharge;
	type = paramType;

	com_x = com_y = com_z = 0;
	analyticR6BR = 0;
	vdwR4Sum = 0;
	vdwR6Sum = 0;
	chunkR6BR = 0;
	chunkVolume = 0;
	molecularR6Sum = 0;
 
        Region = 0;
}

Atom::Atom(const Atom& a) {
	x = a.x; y = a.y; z = a.z;
	origRadius = a.origRadius;
	radius = a.radius;
	number = a.number;
	charge = a.charge;
	type = a.type;
	
	neighbors = a.neighbors;
	chunk = a.chunk;
	
	com_x = a.com_x; com_y = a.com_y; com_z = a.com_z;
	analyticR6BR = a.analyticR6BR;
	vdwR4Sum = a.vdwR4Sum;
	vdwR6Sum = a.vdwR6Sum;
	chunkR6BR = a.chunkR6BR;
	chunkVolume = a.chunkVolume;
	molecularR6Sum = a.molecularR6Sum;

        Region = a.Region;
}

/************************Operator Overloads**********************************/
bool Atom::operator == (const Atom& a) const {
	return number == a.number;
}
ostream& operator << (ostream& out, const Atom& a) {
	out << a.number;
	return out;
}

/**************************Center of Mass************************************/
void Atom::CoM(double paramX, double paramY, double paramZ) {
	com_x = paramX;
	com_y = paramY;
	com_z = paramZ;
}
void Atom::CoM(double &paramX, double &paramY, double &paramZ) const {
	paramX = com_x;
	paramY = com_y;
	paramZ = com_z;
}

/*****************************Distances**************************************/
double Atom::distance(const Atom& a) const {
	return sqrt((x-a.x)*(x-a.x) + (y-a.y)*(y-a.y) + (z-a.z)*(z-a.z));
}
double Atom::distance(double paramX, double paramY, double paramZ) const {
	return sqrt((paramX-x)*(paramX-x) + (paramY-y)*(paramY-y) + (paramZ-z)*(paramZ-z));
}

/*****************************Neighbors**************************************/
bool Atom::addAtomToNeighbors(int atomNumber) {
	pair<set<int>::iterator, bool> p = neighbors.insert(atomNumber);
	return p.second;
}
set<int> Atom::Neighbors() {
	return neighbors;
}
set<int>::iterator Atom::neighborsBegin() {
	return neighbors.begin();
}
set<int>::iterator Atom::neighborsEnd() {
	return neighbors.end();
}
int Atom::neighborsSize() {
	return neighbors.size();
}
void Atom::resetNeighbors() {
	neighbors.clear();
}
bool Atom::isInNeighbors(int atomNumber) {
        set<int>::iterator result = neighbors.find(atomNumber);
        return result != neighbors.end();
}


/********************************Chunks**************************************/
bool Atom::addAtomToChunk(int atomNumber) {
	pair<set<int>::iterator, bool> result = chunk.insert(atomNumber);
	return result.second;
}
bool Atom::addGroupToChunk(const set<int> &group) {
  set<int>::const_iterator i;
  for (i = group.begin(); i != group.end(); i++) {
    //cout << "   Adding " << *i << endl;
  }
	chunk.insert(group.begin(), group.end());
}
bool Atom::removeAtomFromChunk(int atomNumber) {
	return chunk.erase(atomNumber);
}
set<int>::iterator Atom::chunkBegin() {
	return chunk.begin();
}
set<int>::iterator Atom::chunkEnd() {
	return chunk.end();
}
bool Atom::isInChunk(int atomNumber) {
	set<int>::iterator result = chunk.find(atomNumber);
	return result != chunk.end();
}
set<int> Atom::Chunk() {
	return chunk;
}
int Atom::chunkSize() const {
	return chunk.size();
}



/********************************NoNEcks**************************************/
bool Atom::addAtomToNoNeck(int atomNumber) {
        pair<set<int>::iterator, bool> result = noneck.insert(atomNumber);
        return result.second;
}
bool Atom::addGroupToNoNeck(const set<int> &group) {
  set<int>::const_iterator i;
  //for (i = group.begin(); i != group.end(); i++) {
    //cout << "   Adding " << *i << endl;
  //}
        noneck.insert(group.begin(), group.end());
}
bool Atom::removeAtomFromNoNeck(int atomNumber) {
        return noneck.erase(atomNumber);
}
set<int>::iterator Atom::noneckBegin() {
        return noneck.begin();
}
set<int>::iterator Atom::noneckEnd() {
        return noneck.end();
}
bool Atom::isInNoNeck(int atomNumber) {
        set<int>::iterator result = noneck.find(atomNumber);
        return result != noneck.end();
}
set<int> Atom::NoNeck() {
        return noneck;
}
int Atom::NoNeckSize() const {
        return noneck.size();
}




/*******************************Data accessors********************************/
double Atom::X() const {
	return x;
}
double Atom::Y() const {
	return y;
}
double Atom::Z() const {
	return z;
}
double Atom::OrigRadius() const {
	return origRadius;
}
double Atom::Radius() const {
	return radius;
}
int Atom::Number() const {
	return number;
}
double Atom::Charge() const {
	return charge;
}
string Atom::Type() const {
	return type;
}

/**************************Calculation accessors*****************************/
double Atom::AnalyticR6BR() const {
	return analyticR6BR;
}
double Atom::VdwR4Sum() const {
	return vdwR4Sum;
}
double Atom::VdwR6Sum() const {
	return vdwR6Sum;
}
double Atom::ChunkR6BR() const {
	return chunkR6BR;
}
double Atom::ChunkVolume() const {
	return chunkVolume;
}
double Atom::MolecularR6Sum() const {
	return molecularR6Sum;
}
double Atom::ChunkR6Packing() const {
        return chunkR6Packing;
}




/***************************Data modifiers***********************************/
void Atom::X(double paramX) {
	x = paramX;
}
void Atom::Y(double paramY) {
	y = paramY;
}
void Atom::Z(double paramZ) {
	z = paramZ;
}
void Atom::Radius(double paramRadius) {
	radius = paramRadius;
}

/*************************Calculation modifiers******************************/
void Atom::AnalyticR6BR(double value) {
	analyticR6BR = value;
}
void Atom::VdwR4Sum(double value) {
	vdwR4Sum = value;
}
void Atom::VdwR6Sum(double value) {
	vdwR6Sum = value;
}
void Atom::ChunkR6BR(double value) {
	chunkR6BR = value;
}
void Atom::ChunkVolume(double value) {
	chunkVolume = value;
}
void Atom::MolecularR6Sum(double value) {
	molecularR6Sum = value;
}
void Atom::ChunkR6Packing(double value) {
        chunkR6Packing = value;
}




