#include "GBorn.h"


//******************************************************************************

GBorn::GBorn() {
}

// Small function for the mu parameter
double mu_inv( double charge, double Reff, double ROH) {
   double  sign = -1.0 ;
   if ( charge > 0.0 )  sign =  1.0 ;
   else if ( charge == 0.0 ) sign = 0.0;

   double mu_i =  ( 1.0 + sign*ROH/(Reff - 0.52 + 1.4) );
   //double mu_i = (1.0 + sign*0.722/(Reff - 0.52 + 1.4));
   
   return mu_i ;
}

// Canonical GB (Still equation)
double GBorn::gb(Molecule &m, double kappa, double diel_ext, double diel_int ) {

   double  e,epol,egas,dielfac,qi,qiqj,fgb,rb2inv,num; 
   double  epol_offdiagprime, epoldiagprime, sumqiqj;
   double  xi,yi,zi,xij,yij,zij;

   double  dij,sumi;
   double  r2,ri,sj,sj2;
   double  lij,lij2,uij,uij2,efac,efac1,efac2,temp3;
   double  ra2;
   double  e_extra;

   Atoms::iterator i, j;

   /* Compute the GB energy    */
   epol = 0.0; egas = 0.0; e_extra = 0.0; epol_offdiagprime = 0.0, epoldiagprime = 0.0;
   sumqiqj=0.0;

   // These will only affect GB(epsilon). 
   double alpha = 0.57;
   double beta = (diel_int/diel_ext);
   double factor =   1.0/(1.0  + alpha*beta);
   double Rw = 1.4; // water radius 
   double mu = 4.0; // scaling factor 
   double bb = (alpha - 0.5)*beta/( 1 + 0.5*beta);


   for (i = m.begin(); i != m.end(); i++) {
      for (j = i + 1; j != m.end(); j++ ) {
		
         xij = i->X() - j->X();
         yij = i->Y() - j->Y();
         zij = i->Z() - j->Z();
         r2 = xij*xij + yij*yij + zij*zij;
         rb2inv = (i->MolecularR6Sum()) * (j->MolecularR6Sum());//Reffs(eps_ext=Infinity)
		
         efac = exp(-r2*rb2inv/4.0); // FGB. Original. Still
         
         fgb = sqrt(r2 + efac/rb2inv); // FGB Original. Still

         dielfac = 1.0/diel_int - exp(-(kappa)*fgb)/(diel_ext);
         qi = dielfac * (i->Charge() );

         qiqj = qi * (j -> Charge() ) ;

         e = -qiqj / fgb; // GB. Original. Still.  
        
         epol += e;
      }

      dielfac = 1.0/diel_int - exp(-(kappa)/( i->MolecularR6Sum() ))/(diel_ext);

      qiqj = dielfac * (i->Charge() )*(i->Charge() )   ;
      e = -qiqj * ( i->MolecularR6Sum() );

      epol +=  0.5*e;
   }

   return epol*332.05221729;
}

// Charge assymetric GB model 
double GBorn::chagb(Molecule &m, double A, double kappa, double diel_ext, double diel_int , double tau, double ROH) {

   double  e,epol,egas,dielfac,qi,qiqj,fgb, rb2inv2,rb2inv,num; 
   double  epol_offdiagprime, epoldiagprime, sumqiqj;
   double  xi,yi,zi,xij,yij,zij;
   double mu_inv(double charge, double R, double ROH);
   double  dij,sumi;
   double  r2,ri,sj,sj2;
   double  lij,lij2,uij,uij2,efac,efac1,efac2,temp3;
   double  ra2;
   double  e_extra;

   Atoms::iterator i, j;

   /* Compute the GB energy    */
   epol = 0.0; egas = 0.0; e_extra = 0.0; epol_offdiagprime = 0.0, epoldiagprime = 0.0;
   sumqiqj=0.0;

   // These will only affect GB(epsilon). 
   double alpha = 0.57;
   double beta = (diel_int/diel_ext);
   double factor =   1.0/(1.0  + alpha*beta);
   double ab_over_A = alpha*beta/A; 

   // computation of net charge for each atom (qreff) 
   for (i = m.begin(); i != m.end(); i++) {
      i->qeff = 0;

      for (j = m.begin(); j != m.end(); j++) {

          xij = i->X() - j->X();
          yij = i->Y() - j->Y();
          zij = i->Z() - j->Z();

          r2 = xij*xij + yij*yij + zij*zij;

          i->qeff += (j-> Charge())*exp(-(tau*r2*i->MolecularR6Sum()*j->MolecularR6Sum()));//Boris! 
      }
   }

   // Computen Solvation energies
   for (i = m.begin(); i != m.end(); i++) {
      for (j = i + 1; j != m.end(); j++ ) {

         xij = i->X() - j->X();
         yij = i->Y() - j->Y();
         zij = i->Z() - j->Z();

	 r2 = xij*xij + yij*yij + zij*zij;
         rb2inv = (i->MolecularR6Sum())*(j->MolecularR6Sum()/(mu_inv( i->qeff, 1/i->MolecularR6Sum() , ROH)*mu_inv( j->qeff, 1/j->MolecularR6Sum() , ROH) ) ); // Reffs(eps_ext=Infinity)
         rb2inv2=i->MolecularR6Sum() * j-> MolecularR6Sum();      
         efac = exp(-r2*rb2inv2/4.0); // FGB. Original. Still

         fgb = sqrt(r2 + efac/rb2inv); // FGB Original. Still

         dielfac = 1.0/diel_int - exp(-(kappa)*fgb)/(diel_ext);

         qi = dielfac * (i->Charge() );

         qiqj = qi * (j -> Charge() ) ;

         //e = -qiqj / fgb; // GB. Original. Still.  
         //A = A*(1.0 + kappa*A); // salt correction 
         e = -qiqj * ( 1.0/fgb + ab_over_A ); // ALPB  
        
         epol += e;
      }

      dielfac = 1.0/diel_int - exp(-(kappa)/( i->MolecularR6Sum() ))/(diel_ext);

      qiqj = dielfac*(i->Charge())*(i->Charge()) ;
      e= -qiqj*(i->MolecularR6Sum()/mu_inv(i->qeff,1/i->MolecularR6Sum(),ROH)+ab_over_A);

      epol +=  0.5*e;


   }
   // rescale the energy to give [kcal/mol] if distance is in Angstroms
   // E = qiqj/rij
   // 332.02221729 = 18.2223*18.2223 

   return epol*factor*332.05221729;
}
// Charge assymetric GB model 
double GBorn::chagbcan(Molecule &m, double kappa, double diel_ext, double diel_int , double tau, double ROH) {

   double  e,epol,egas,dielfac,qi,qiqj,fgb, rb2inv2,rb2inv,num; 
   double  epol_offdiagprime, epoldiagprime, sumqiqj;
   double  xi,yi,zi,xij,yij,zij;
   double mu_inv(double charge, double R, double ROH);
   double  dij,sumi;
   double  r2,ri,sj,sj2;
   double  lij,lij2,uij,uij2,efac,efac1,efac2,temp3;
   double  ra2;
   double  e_extra;

   Atoms::iterator i, j;

   /* Compute the GB energy    */
   epol = 0.0; egas = 0.0; e_extra = 0.0; epol_offdiagprime = 0.0, epoldiagprime = 0.0;
   sumqiqj=0.0;

   // computation of net charge for each atom (qreff) 
   for (i = m.begin(); i != m.end(); i++) {
      i->qeff = 0;

      for (j = m.begin(); j != m.end(); j++) {

          xij = i->X() - j->X();
          yij = i->Y() - j->Y();
          zij = i->Z() - j->Z();

          r2 = xij*xij + yij*yij + zij*zij;

          i->qeff += (j-> Charge())*exp(-(tau*r2*i->MolecularR6Sum()*j->MolecularR6Sum()));//Boris! 
      }
   }

   // Computen Solvation energies
   for (i = m.begin(); i != m.end(); i++) {
      for (j = i + 1; j != m.end(); j++ ) {

         xij = i->X() - j->X();
         yij = i->Y() - j->Y();
         zij = i->Z() - j->Z();

	 r2 = xij*xij + yij*yij + zij*zij;
         rb2inv = (i->MolecularR6Sum())*(j->MolecularR6Sum()/(mu_inv( i->qeff, 1/i->MolecularR6Sum() , ROH)*mu_inv( j->qeff, 1/j->MolecularR6Sum() , ROH) ) ); // Reffs(eps_ext=Infinity)
         rb2inv2=i->MolecularR6Sum() * j-> MolecularR6Sum();      
         efac = exp(-r2*rb2inv2/4.0); // FGB. Original. Still

         fgb = sqrt(r2 + efac/rb2inv); // FGB Original. Still

         dielfac = 1.0/diel_int - exp(-(kappa)*fgb)/(diel_ext);

         qi = dielfac * (i->Charge() );

         qiqj = qi * (j -> Charge() ) ;

         //e = -qiqj / fgb; // GB. Original. Still.  
         //A = A*(1.0 + kappa*A); // salt correction 
         e = -qiqj/fgb; // ALPB  
        
         epol += e;
      }

      dielfac = 1.0/diel_int - exp(-(kappa)/( i->MolecularR6Sum() ))/(diel_ext);

      qiqj = dielfac*(i->Charge())*(i->Charge()) ;
      e= -qiqj*(i->MolecularR6Sum()/mu_inv(i->qeff,1/i->MolecularR6Sum(),ROH));

      epol +=  0.5*e;


   }
   // rescale the energy to give [kcal/mol] if distance is in Angstroms
   // E = qiqj/rij
   // 332.02221729 = 18.2223*18.2223 

   return epol*332.05221729;
}


double GBorn::alpb(Molecule &m, double A, double kappa, double diel_ext, double diel_int ) {

	double  e,epol,egas,dielfac,qi,qiqj,fgb,rb2inv,num, epol_offdiagprime, epoldiagprime, sumqiqj;
	double  xi,yi,zi,xij,yij,zij;

	double  dij,sumi;
	double  r2,ri,sj,sj2;
	double  lij,lij2,uij,uij2,efac,efac1,efac2,temp3;
	double  ra2;
	double  e_extra;


	Atoms::iterator i, j;


	/*    Compute the GB energy    */
	epol = 0.0; egas = 0.0; e_extra = 0.0; epol_offdiagprime = 0.0, epoldiagprime = 0.0;
	sumqiqj=0.0;

	// These will only affect GB(epsilon). 
	double alpha = 0.57;
	double beta = (diel_int/diel_ext);
	double factor =   1.0/(1.0  + alpha*beta);
        double ab_over_A = alpha*beta/A; 


	for (i = m.begin(); i != m.end(); i++) {

			
	   	for (j = i + 1; j != m.end(); j++ ) {
		
              		xij = i->X() - j->X();
              		yij = i->Y() - j->Y();
              		zij = i->Z() - j->Z();
              		r2 = xij*xij + yij*yij + zij*zij;
              		rb2inv = ( i->MolecularR6Sum()  ) * (j-> MolecularR6Sum() ) ; // Reffs(eps_ext=Infinity)
		
	      		efac = exp(-r2*rb2inv/4.0); // FGB. Original. Still

	      		fgb = sqrt(r2 + efac/rb2inv); // FGB Original. Still

                	dielfac = 1.0/diel_int - exp(-(kappa)*fgb)/(diel_ext);
                	qi = dielfac * (i->Charge() );

                	qiqj = qi * (j -> Charge() ) ;

                	//e = -qiqj / fgb; // GB. Original. Still.  
                        //A = A*(1.0 + kappa*A); // salt correction 
                	e = -qiqj * ( 1.0/fgb + ab_over_A ); // ALPB  
        
                	epol += e;

		}

		dielfac = 1.0/diel_int - exp(-(kappa)/( i->MolecularR6Sum() ))/(diel_ext);

		qiqj = dielfac * (i->Charge() )*(i->Charge() )   ;
		e = -qiqj * ( i->MolecularR6Sum() +  ab_over_A);

        	epol +=  0.5*e;


	}

	// rescale the energy to give [kcal/mol] if distance is in Angstroms
	// E = qiqj/rij
	// 332.02221729 = 18.2223*18.2223 
	return epol*factor*332.05221729;

}

double GBorn::GetSASFromFile(string msmsFile) {

        // Read the SAS Area
        double SASarea = -1.0;
        string FirstWord="";
        ifstream In(msmsFile.c_str());
        if (!In) {
                cerr << "Error opening msms file pqr " << msmsFile << endl;
                return -1.0;
        }

        while (In && !In.eof()) {
                In >> FirstWord;
                if (FirstWord == "SAS"  ) {
                        In  >> SASarea >> SASarea;
                        In  >> SASarea >> SASarea;
                        In  >> SASarea >> SASarea;
                        In  >> SASarea;
                        break;
                }
        }

        if (FirstWord != "SAS" ) {
                cerr << " SAS not found in msmsFile: " << msmsFile << endl;
                return -1.0;
        }


        In.close();
        return SASarea;

}

double GBorn::SASA( Molecule &m, double ProbeRadius ) {

        int debug = 0;
        int d_junk;
        const char *c_junk;
        double SASarea;

        Atoms::iterator itr;
        FILE *temp = fopen("tempOut", "w");
        for (itr = m.begin(); itr != m.end(); itr++) {
                fprintf(temp, "%f %f %f %f\n",
                 itr->X(), itr->Y(), itr->Z(), itr->Radius());
        }
        fclose(temp);


	//Covert the Probe Radius from double to string 
        std::ostringstream oss;
        oss << ProbeRadius;
        std::string Rw = oss.str();

        //--------------<HACK>--------------
        //Run msms on the file from the command line
        string cmd = "msms -if tempOut -prob " + Rw + " -surface ases > my_msms.out 2> /dev/null";
        system(cmd.c_str());
        //--------------</HACK>-------------

        SASarea = GetSASFromFile("my_msms.out");
        if (SASarea == -1.0) {
        	cerr << "possible problems running msms with your structure " << endl;
                return -1.0;
        }
        
	return SASarea;
	//cout << gamma * SASarea << endl;

}



double GBorn::VDWenergy( Molecule &m, double ProbeRadius , string ParameterFile ) {

 //double reff[MAX]; // Effective radii 
 string AtomType; // used to store the atom name
 //int AtomIndexes[MAX]; // stores every  index of the Atoms of the pqr file;

 int i,j, num;
 int natom = 0 ; //number of atoms obtained from the pqr file
 int natom1 = 0; // number of atoms obtained from the Reff file

 double R_w = ProbeRadius; // Water probe 
 double d_w = 0.033428; // number density of water at standard conditions
 double sigma_w = 3.1507; // Water Lennar Jones
 double epsilon_w = 0.1520; // Parameters of tip3p Water

 double a_i = 0.0;// Parameter used in equation (28) of the GB_CHUNK paper
 double sigma_i = 0.0; //Lennar JOnes parameters for each atom
 double epsilon_i = 0.0;
 double sigma_iw = 0.0;
 double epsilon_iw = 0.0;
 double mu_i = 1.0; // Adjustable paremeters, see equation 28 in GB_CHUNKS paper  


 // Build one Table (Types) to store 
 // the index used for the diferent atom types (two letters) 
 // especified in the Lennard Jones parameters file 
 int Types[TABLEX][TABLEY];
 double Sigma[ATOM_TYPES + 1];
 double Epsilon[ATOM_TYPES + 1];
 double Mu[ATOM_TYPES + 1];
 // initializing the values in 0
 for (i = 0; i < TABLEX ; i++  ){
        for (j = 0; j < TABLEY ; j++){
                Types[i][j] = 0;
        }
 }

 for (i = 0; i < ATOM_TYPES + 1 ; i++){
        Sigma[i] = 0.0;
        Epsilon[i] = 0.0;
        Mu[i] = 0.0 ;
 }

 fillTables(Types, Sigma, Epsilon, Mu, ParameterFile );  // Filling the Tables 

 string junk;
 Atoms::iterator ii, jj;
 for (ii = m.begin(); ii != m.end(); ii++) {

 	junk = ii->Type();
	if ( isdigit(junk[0]) )  AtomType = junk.erase(0,1);
        else AtomType = junk;


        if ( (int)AtomType[0] < 65 || (int)AtomType[0] > 90) {
        	cerr << "Error reading atom name :" << AtomType <<  endl ;
                exit(1);
        }
        // Finding the corresponding Index for the corresponding
        // Atomytpe
        if ( AtomType.length() == 1){
                i = (int) AtomType[0] - 65;
                j = 0;
        }
        else {
                i = (int) AtomType[0] - 65;
                j = (int) AtomType[1] - 47;  //Table begin in ascii of "0" = 47
                if (j < 0 || j > TABLEY) j = 91-47; // It wll take ascci place of "["
        }

        ii -> VDWindex  =  Types[i][j];

 }

 // Loop over the atoms to compute the a_i , and the e_vdw
 // a_i : parameter obatined by the Lennard Jones parameters see, Galliccio et. al.
 // e_vdw : Solvention energy described in Galliccio et al/
 double e_vdw = 0.0;
 double reff =0.0;
 for ( ii = m.begin(); ii != m.end(); ii++) {
        //computation of a_i (see GB_CHUNK paper)
        sigma_i = Sigma[ ii->VDWindex  ];
        epsilon_i = Epsilon[ ii->VDWindex ];
        mu_i = Mu[ ii->VDWindex ];

        sigma_iw =0.25* (sigma_i + sigma_w )*(sigma_i + sigma_w );
	epsilon_iw = sqrt(epsilon_i * epsilon_w);

        a_i =  - 16 * M_PI * d_w * epsilon_iw * sigma_iw * sigma_iw * sigma_iw / 3;
        //a_i =  d_w * epsilon_iw * sigma_iw * sigma_iw * sigma_iw ;

        reff = 1/( ii->MolecularR6Sum() );
        e_vdw = e_vdw + mu_i*a_i /( (reff+R_w )*(reff+R_w)*(reff+R_w) );

        //cout << sigma_i << "   " << epsilon_i << "  " << a_i << "   " << mu_i << endl;

 }
 
 return e_vdw;

}
//////////////////////////////////////////////////////
// Purpose, Fill the table "Types" which store indexes for each type of atoms
//  especified by LennardJones.dat File. 
void GBorn::fillTables(int Types[TABLEX][TABLEY], double  Sigma[ATOM_TYPES] , double Epsilon[ATOM_TYPES], double Mu[ATOM_TYPES] , string ParameterFile ) { // Filling the Tables 
        
        int i,j;
        string AtomType;
        int AtomIndex = 0;

        // Reading the values of sigma and epsilon form the file LennardJones.dat
        // and saving it in their respecitve tables
        // string ParameterFile = "LennardJones.dat";
  
        
  
        ifstream IN;

        IN.open(ParameterFile.c_str());
        if (!IN) {
                cerr << "Error reading " << ParameterFile << "  parameter file "<< endl;
                exit(1);
        }


        IN >> AtomType;
        while ( IN  ) {

                if ( (int)AtomType[0] < 65 || (int)AtomType[0] > 90) {
                        cerr << "Error reading atom name in Lenna Jones file" << endl ;
                        exit(1);
                }

                AtomIndex++;

                if ( AtomType.size() == 1) {
                        i = (int)AtomType[0] - 65;
                        for (j=0 ; j < TABLEY ;  j++) {
                                Types[i][j] = AtomIndex ;
                        }
                }
                else {
                        i = (int)AtomType[0] - 65;
                        j = (int)AtomType[1] - 47;
                        if (j < 0 || j > TABLEY) j = 91-47; // It will take the ascii place of '['
                        Types[i][j] = AtomIndex ;
                }


                IN >> Sigma[AtomIndex] >> Epsilon[AtomIndex] >> Mu[AtomIndex]  ;

                IN.ignore(255, '\n');
                 
               IN >> AtomType;
        }
        IN.clear();
        IN.close();

        AtomIndex = 0;

        IN.open(ParameterFile.c_str());
        if (!IN) {
                cerr << "Error reading the LennarJones parameter file "<< endl;
                exit(1);
        }

        IN >> AtomType;
        while ( IN  ) {

                if ( (int)AtomType[0] < 65 || (int)AtomType[0] > 90) {
                        cerr << "Error reading atom name in Lenna Jones file" << endl ;
                        exit(1);
                }

                AtomIndex++;

                if ( AtomType.size() == 1) {
                        i = (int)AtomType[0] - 65;
                        j = 0;
                }
                else {
                        i = (int)AtomType[0] - 65;
                        j = (int)AtomType[1] - 47;
                        if (j < 0 || j > TABLEY) j = 91-47; // It wll take ascci place of "["
                }
                Types[i][j] = AtomIndex;

                IN.ignore(255, '\n');
                IN >> AtomType;
        }
        IN.clear();
        IN.close();

}

//  Coulomb energy 
double GBorn::Coulomb(Molecule &m, double diel ) {

	double  e,ecoul;
	double  xij,yij,zij;
	double  qi,qj,qiqj;
	double  r2,rij;
	
	Atoms::iterator i, j;

	/* Compute coulomb energy    */
	ecoul = 0.0; 

	for (i = m.begin(); i != m.end(); i++) {
		
		qi = i->Charge(); 	
		
	   	for (j = i + 1; j != m.end(); j++ ) {
		
              		xij = i->X() - j->X();
              		yij = i->Y() - j->Y();
              		zij = i->Z() - j->Z();
              	
			r2 = xij*xij + yij*yij + zij*zij;
			
			rij = sqrt(r2);

                	qiqj = qi * (j -> Charge() ) ;

                	e = qiqj / rij; // Coulomb energy .  
        
                	ecoul += e;

		}
	}
	// rescale the energy to give [kcal/mol] if distance is in Angstroms
	// E = qiqj/rij
	// 332.02221729 = 18.2223*18.2223 
	return ecoul*332.05221729/diel;
}
double GBorn::gb_ij(Molecule &m, double kappa, double diel_ext, double diel_int, string DGijFile, bool solv_only) {

  double  e,epol,egas,dielfac,qi,qiqj,fgb,rb2inv,num, epol_offdiagprime, epoldiagprime, sumqiqj;
  double  xi,yi,zi,xij,yij,zij;
  double  ecol=0.0;
  double  dij,sumi;
  double  r2,ri,sj,sj2, rij;
  double  lij,lij2,uij,uij2,efac,efac1,efac2,temp3;
  double  ra2;
  double  e_extra;

  Atoms::iterator i, j;
  /* Set up output file  */
  ofstream out(DGijFile.c_str());
  if (!out) 
      cerr << "problems setting up " << DGijFile  << " as output file"  << endl ;
       

   /*Compute the GB energy    */
   epol = 0.0; egas = 0.0; e_extra = 0.0; epol_offdiagprime = 0.0, epoldiagprime = 0.0;
   sumqiqj=0.0;

   // These will only affect GB(epsilon). 
   for (i = m.begin(); i != m.end(); i++) {

      dielfac = 1.0/diel_int - exp(-(kappa)/( i->MolecularR6Sum() ))/(diel_ext);

      qiqj = dielfac * (i->Charge() )*(i->Charge() )   ;
      e = -qiqj * ( i->MolecularR6Sum());

      out<<i->Number()<<" "<<i->Number()<<" "<<0.5*332.05221729*e << endl;
      epol += 0.5*e;

      for (j = i + 1; j != m.end(); j++ ) {

         xij = i->X() - j->X();
         yij = i->Y() - j->Y();
         zij = i->Z() - j->Z();
         r2 = xij*xij + yij*yij + zij*zij;
         rij = sqrt(r2);

         // Reffs(eps_ext=Infinity)
         rb2inv = ( i->MolecularR6Sum()  ) * (j-> MolecularR6Sum() ) ; 

         efac = exp(-r2*rb2inv/4.0); // FGB. Original. Still

         fgb = sqrt(r2 + efac/rb2inv); // FGB Original. Still

         dielfac = 1.0/diel_int - exp(-(kappa)*fgb)/(diel_ext);
         qi = dielfac * (i->Charge() );

         qiqj = qi * (j -> Charge() ) ;

         //e = -qiqj / fgb; // GB. Original. Still.  
         //A = A*(1.0 + kappa*A); // salt correction 
         e = -qiqj * ( 1.0/fgb ); // ALPB

         if (solv_only)
            out<<i->Number()<<" "<<j->Number()<<" "<<332.05221729*e<<endl;
         else {
            ecol = (i->Charge() )*(j -> Charge()) / rij; // Coulomb energy
            out << i->Number() << " " << j->Number() << " " <<  332.05221729*e + ecol*332.05221729/diel_int  << endl;
         }


         epol += e;

      }
   }

   // rescale the energy to give [kcal/mol] if distance is in Angstroms
   // E = qiqj/rij
   // 332.02221729 = 18.2223*18.2223 
   return epol*332.05221729;

}
double GBorn::alpb_ij(Molecule &m, double A, double kappa, double diel_ext, double diel_int, string DGijFile, double alpha_alpb, bool solv_only) {

  double  e,epol,egas,dielfac,qi,qiqj,fgb,rb2inv,num, epol_offdiagprime, epoldiagprime, sumqiqj;
  double  xi,yi,zi,xij,yij,zij;
  double  ecol=0.0;
  double  dij,sumi;
  double  r2,ri,sj,sj2, rij;
  double  lij,lij2,uij,uij2,efac,efac1,efac2,temp3;
  double  ra2;
  double  e_extra;

  Atoms::iterator i, j;
  /* Set up output file  */
  ofstream out(DGijFile.c_str());
  if (!out) 
      cerr << "problems setting up " << DGijFile  << " as output file"  << endl ;
       

   /*Compute the GB energy    */
   epol = 0.0; egas = 0.0; e_extra = 0.0; epol_offdiagprime = 0.0, epoldiagprime = 0.0;
   sumqiqj=0.0;

   // These will only affect GB(epsilon). 
   double alpha = 0.57;
   double beta = (diel_int/diel_ext);
   double factor =   1.0/(1.0  + alpha*beta);
   double ab_over_A = alpha*beta/A;

   for (i = m.begin(); i != m.end(); i++) {

      dielfac = 1.0/diel_int - exp(-(kappa)/( i->MolecularR6Sum() ))/(diel_ext);

      qiqj = dielfac * (i->Charge() )*(i->Charge() )   ;
      e = -qiqj * ( i->MolecularR6Sum() +  ab_over_A);

      out<<i->Number()<<" "<<i->Number()<<" "<<0.5*factor*332.05221729*e << endl;
      epol += 0.5*e;

      for (j = i + 1; j != m.end(); j++ ) {

         xij = i->X() - j->X();
         yij = i->Y() - j->Y();
         zij = i->Z() - j->Z();
         r2 = xij*xij + yij*yij + zij*zij;
         rij = sqrt(r2);

         // Reffs(eps_ext=Infinity)
         rb2inv = ( i->MolecularR6Sum()  ) * (j-> MolecularR6Sum() ) ; 

         efac = exp(-r2*rb2inv/4.0); // FGB. Original. Still

         fgb = sqrt(r2 + efac/rb2inv); // FGB Original. Still

         dielfac = 1.0/diel_int - exp(-(kappa)*fgb)/(diel_ext);
         qi = dielfac * (i->Charge() );

         qiqj = qi * (j -> Charge() ) ;

         //e = -qiqj / fgb; // GB. Original. Still.  
         //A = A*(1.0 + kappa*A); // salt correction 
         e = -qiqj * ( 1.0/fgb + ab_over_A ); // ALPB

         if (solv_only)
            out<<i->Number()<<" "<<j->Number()<<" "<<factor*332.05221729*e<<endl;
         else {
            ecol = (i->Charge() )*(j -> Charge()) / rij; // Coulomb energy
            out << i->Number() << " " << j->Number() << " " <<  factor*332.05221729*e + ecol*332.05221729/diel_int  << endl;
         }


         epol += e;

      }
   }

   // rescale the energy to give [kcal/mol] if distance is in Angstroms
   // E = qiqj/rij
   // 332.02221729 = 18.2223*18.2223 
   return epol*factor*332.05221729;

}

// Charge assymetric GB model 
double GBorn::chagb_ij(Molecule &m, double A, double kappa, double diel_ext, double diel_int , double tau, double ROH , string DGijFile , bool solv_only ) {

   double  e,epol,egas,dielfac,qi,qiqj,fgb, rb2inv2,rb2inv,num;
   double  epol_offdiagprime, epoldiagprime, sumqiqj;
   double  xi,yi,zi,xij,yij,zij;
   double mu_inv(double charge, double R, double ROH);
   double  dij,sumi;
   double  r2,ri,sj,sj2;
   double  lij,lij2,uij,uij2,efac,efac1,efac2,temp3;
   double  ra2;
   double  e_extra;

   Atoms::iterator i, j;

   /* Set up output file  */
   ofstream out(DGijFile.c_str());
   if (!out) {
       cerr << "problems setting up " << DGijFile  << " as output file"  << endl ;
   }

   /* Compute the GB energy    */
   epol = 0.0; egas = 0.0; e_extra = 0.0; epol_offdiagprime = 0.0, epoldiagprime = 0.0;
   sumqiqj=0.0;

   // These will only affect GB(epsilon). 
   double alpha = 0.57;
   double beta = (diel_int/diel_ext);
   double factor =   1.0/(1.0  + alpha*beta);
   double ab_over_A = alpha*beta/A;

   // computation of net charge for each atom (qreff) 
   for (i = m.begin(); i != m.end(); i++) {
      i->qeff = 0;

      for (j = m.begin(); j != m.end(); j++) {

          xij = i->X() - j->X();
          yij = i->Y() - j->Y();
          zij = i->Z() - j->Z();

          r2 = xij*xij + yij*yij + zij*zij;

          i->qeff += (j-> Charge())*exp(-(tau*r2*i->MolecularR6Sum()*j->MolecularR6Sum()));//Boris! 
      }
   }

   // Computen Solvation energies
   for (i = m.begin(); i != m.end(); i++) {

      dielfac = 1.0/diel_int - exp(-(kappa)/( i->MolecularR6Sum() ))/(diel_ext);
      qiqj = dielfac*(i->Charge())*(i->Charge()) ;
      e= -qiqj*(i->MolecularR6Sum()/mu_inv(i->qeff,1/i->MolecularR6Sum(),ROH)+ab_over_A);
      out << i->Number() << " " << i->Number() << " " <<  0.5*factor*332.05221729*e << endl;
      epol +=  0.5*e;

      for (j = i + 1; j != m.end(); j++ ) {

         xij = i->X() - j->X();
         yij = i->Y() - j->Y();
         zij = i->Z() - j->Z();

         r2 = xij*xij + yij*yij + zij*zij;
         double rij = sqrt(r2);
         rb2inv = (i->MolecularR6Sum())*(j->MolecularR6Sum()/(mu_inv( i->qeff, 1/i->MolecularR6Sum() , ROH)*mu_inv( j->qeff, 1/j->MolecularR6Sum() , ROH) ) ); // Reffs(eps_ext=Infinity)
         rb2inv2=i->MolecularR6Sum() * j-> MolecularR6Sum();
         efac = exp(-r2*rb2inv2/4.0); // FGB. Original. Still

         fgb = sqrt(r2 + efac/rb2inv); // FGB Original. Still

         dielfac = 1.0/diel_int - exp(-(kappa)*fgb)/(diel_ext);

         qi = dielfac * (i->Charge() );

         qiqj = qi * (j -> Charge() ) ;

         //e = -qiqj / fgb; // GB. Original. Still.  
         //A = A*(1.0 + kappa*A); // salt correction 
         e = -qiqj * ( 1.0/fgb + ab_over_A ); // ALPB  

         if (solv_only) {
            out << i->Number() << " " << j->Number() << " " <<  factor*332.05221729*e << endl;
         }
         else {
            double ecol = (i->Charge() )*(j -> Charge()) / rij; // Coulomb energy
            out << i->Number() << " " << j->Number() << " " <<  factor*332.05221729*e + ecol*332.05221729/diel_int  << endl;
         }

         epol += e;
      }
   }
   // rescale the energy to give [kcal/mol] if distance is in Angstroms
   // E = qiqj/rij
   // 332.02221729 = 18.2223*18.2223 

   return epol*factor*332.05221729;
}

// Charge assymetric GB model 
double GBorn::chagbcan_ij(Molecule &m, double kappa, double diel_ext, double diel_int , double tau, double ROH , string DGijFile , bool solv_only ) {

   double  e,epol,egas,dielfac,qi,qiqj,fgb, rb2inv2,rb2inv,num;
   double  epol_offdiagprime, epoldiagprime, sumqiqj;
   double  xi,yi,zi,xij,yij,zij;
   double mu_inv(double charge, double R, double ROH);
   double  dij,sumi;
   double  r2,ri,sj,sj2;
   double  lij,lij2,uij,uij2,efac,efac1,efac2,temp3;
   double  ra2;
   double  e_extra;

   Atoms::iterator i, j;

   /* Set up output file  */
   ofstream out(DGijFile.c_str());
   if (!out) {
       cerr << "problems setting up " << DGijFile  << " as output file"  << endl ;
   }

   /* Compute the GB energy    */
   epol = 0.0; egas = 0.0; e_extra = 0.0; epol_offdiagprime = 0.0, epoldiagprime = 0.0;
   sumqiqj=0.0;


   // computation of net charge for each atom (qreff) 
   for (i = m.begin(); i != m.end(); i++) {
      i->qeff = 0;

      for (j = m.begin(); j != m.end(); j++) {

          xij = i->X() - j->X();
          yij = i->Y() - j->Y();
          zij = i->Z() - j->Z();

          r2 = xij*xij + yij*yij + zij*zij;

          i->qeff += (j-> Charge())*exp(-(tau*r2*i->MolecularR6Sum()*j->MolecularR6Sum()));//Boris! 
      }
   }

   // Computen Solvation energies
   for (i = m.begin(); i != m.end(); i++) {

      dielfac = 1.0/diel_int - exp(-(kappa)/( i->MolecularR6Sum() ))/(diel_ext);
      qiqj = dielfac*(i->Charge())*(i->Charge()) ;
      e= -qiqj*(i->MolecularR6Sum()/mu_inv(i->qeff,1/i->MolecularR6Sum(),ROH));
      out << i->Number() << " " << i->Number() << " " <<  0.5*332.05221729*e << endl;
      epol +=  0.5*e;

      for (j = i + 1; j != m.end(); j++ ) {

         xij = i->X() - j->X();
         yij = i->Y() - j->Y();
         zij = i->Z() - j->Z();

         r2 = xij*xij + yij*yij + zij*zij;
         double rij = sqrt(r2);
         rb2inv = (i->MolecularR6Sum())*(j->MolecularR6Sum()/(mu_inv( i->qeff, 1/i->MolecularR6Sum() , ROH)*mu_inv( j->qeff, 1/j->MolecularR6Sum() , ROH) ) ); // Reffs(eps_ext=Infinity)
         rb2inv2=i->MolecularR6Sum() * j-> MolecularR6Sum();
         efac = exp(-r2*rb2inv2/4.0); // FGB. Original. Still

         fgb = sqrt(r2 + efac/rb2inv); // FGB Original. Still

         dielfac = 1.0/diel_int - exp(-(kappa)*fgb)/(diel_ext);

         qi = dielfac * (i->Charge() );

         qiqj = qi * (j -> Charge() ) ;

         //e = -qiqj / fgb; // GB. Original. Still.  
         //A = A*(1.0 + kappa*A); // salt correction 
         e = -qiqj/fgb; // ALPB  

         if (solv_only) {
            out << i->Number() << " " << j->Number() << " " <<  332.05221729*e << endl;
         }
         else {
            double ecol = (i->Charge() )*(j -> Charge()) / rij; // Coulomb energy
            out << i->Number() << " " << j->Number() << " " <<  332.05221729*e + ecol*332.05221729/diel_int  << endl;
         }

         epol += e;
      }
   }
   // rescale the energy to give [kcal/mol] if distance is in Angstroms
   // E = qiqj/rij
   // 332.02221729 = 18.2223*18.2223 

   return epol*332.05221729;
}


