/*****************************************************************************
//Created  by Boris Aguilar 08/12/09
//
//
//Description:
//Calculates Electrostativ energy using the generalize born model or ALPB
//radii for the chunk through msms, then pairwise descreening with necks.
//
*****************************************************************************/


#ifndef GBORN_H
#define GBORN_H

#include <math.h>
#include <vector>
#include <stdlib.h>
#include <sstream>

using namespace std;

#include "Molecule.h"

#define MAX 10000
#define S 8.0
#define TABLEX 26
#define TABLEY 76
#define ATOM_TYPES 100


class GBorn {
public:
 	GBorn();

        // this function is used to handle assymetri 
        static double gb(Molecule &m, double kappa=0, double diel_ext=80.0, double diel_int=1.0 );
	
	static double chagb(Molecule &m, double A, double kappa=0, double diel_ext=80.0, double diel_int=1.0, double tau=1.47, double ROH =0.58588 );
	static double chagbcan(Molecule &m, double kappa=0, double diel_ext=80.0, double diel_int=1.0, double tau=1.47, double ROH =0.58588 );
    static double alpb(Molecule &m, double A, double kappa=0, double diel_ext=80.0, double diel_int=1.0);
	static double GetSASFromFile(string msmsFile);
	static double SASA(Molecule &m, double Rw);
    static double VDWenergy( Molecule &m, double ProbeRadius, string ParameterFile ); 
	static void fillTables(int Types[TABLEX][TABLEY], double  Sigma[ATOM_TYPES] , double Epsilon[ATOM_TYPES], double Mu[ATOM_TYPES], string ParamterFile   ) ;
	static double Coulomb(Molecule &m, double diel_int );
	static double alpb_ij(Molecule &m, double A, double kappa=0, double diel_ext=80.0, double diel_int=1.0, string DGijFile="", double alpha_alpb = 0.57, bool solv_only = false );
 	static double gb_ij(Molecule &m, double kappa=0, double diel_ext=80.0, double diel_int=1.0, string DGijFile="", bool solv_only = false );
    static double chagbcan_ij(Molecule &m, double kappa=0, double diel_ext=80.8, double diel_int=1.0 , double tau=1.47 , double ROH=0.58588, string DGijFile="" , bool solv_only=false ); 
    static double chagb_ij(Molecule &m, double A, double kappa=0, double diel_ext=80.8, double diel_int=1.0 , double tau=1.47 , double ROH=0.58588, string DGijFile="" , bool solv_only=false ); 

private:

};

#endif
