#include "MolecularIntegration.h"

float epsin = 1.;
float epsout = 1000.;
// the last coefficient is for 1/(min. dist. to surface)
float BornCoeffients[NUMBORNCOEF];
/******************************************************************************/
int MolecularIntegration::SingleAtomR6(Molecule &m, int atomN, 
 int density, double ProbeRadius ) {
	int d_junk;
	const char *c_junk;
	// checks if the MSMS files exist, assigns file handlers
	Atoms::iterator itr;
	FILE *temp = fopen("tempOut", "w");
	for (itr = m.begin(); itr != m.end(); itr++) {
		fprintf(temp, "%f %f %f %f\n", itr->X(), itr->Y(), itr->Z(),
		 itr->Radius());
	}
	fclose(temp);
	
	//Converts density to a string
	string d;
	stringstream ConvertStream;
	ConvertStream << density;
	ConvertStream >> d;

        //Covert the Probe Radius from double to string 
        std::ostringstream oss;
        oss << ProbeRadius;
        std::string Rw = oss.str();

	//--------------<HACK>--------------
	//Run msms on the file from the command line
	string cmd = "msms -if tempOut -prob " + Rw + " -density " + d +
	 " -no_header -of tempIn 1> /dev/null 2> /dev/null";
	system(cmd.c_str());
	//--------------</HACK>-------------
	
	msms_t *msms = (msms_t*) open_file_read("tempIn", c_junk, &d_junk );
	if ( msms == NULL ) {
		cerr << "Error opening MSMS file" << endl; 
		return 1;
	}

	int   facecount = 0;
	int   vertexcount = 0;
	int   *facet;
	float *vertex, *normal;
	int i, j, t;
	unsigned addr;
	float tf;
	
	// count number of faces
	while (fscanf(msms->ffd, "%d %d %d %d %d", &t, &t, &t, &t, &t) == 5) 
		facecount++;
  	rewind(msms->ffd);
	
	// count number of vertices
	while (fscanf(msms->vfd, "%f %f %f %f %f %f %d %d %d", 
         &tf, &tf, &tf, &tf, &tf, &tf, &t, &t, &t) == 9)
		vertexcount++;
	rewind(msms->vfd);

	// simple sanity check to insure we have at least one usable triangle
	if (facecount < 1 || vertexcount < 3) {
   		fprintf(stderr, "Insufficient data facecount=%d vertexcount=%d",
		 facecount, vertexcount);
		return 1;
	}

	// allocate storage for vertex and normal data
	vertex = new float[3 * vertexcount];
	normal = new float[3 * vertexcount];
	facet  = new int[3 * facecount];
  
	// read in the vertex data
	for (i=0; i<vertexcount; i++) {
		addr = i * 3;
		//Not used
		int atomid, l0fa, l;
		fscanf(msms->vfd, "%f %f %f %f %f %f %d %d %d",
		 &vertex[addr], &vertex[addr+1], &vertex[addr+2], 
		 &normal[addr], &normal[addr+1], &normal[addr+2],
		 &l0fa, &atomid, &l);
	}

	// read in the facet data
	for (i=0; i<facecount; i++) {
		int v0, v1, v2, surftype, ana;
		//read in the next facet
		fscanf(msms->ffd, "%d %d %d %d %d",
		 &v0, &v1, &v2, &surftype, &ana);
		// convert from 1-based indexing to 0-based indexing
    		v0--; v1--; v2--;
    		addr = i * 3;
		facet[addr] = v0;
		facet[addr+1] = v1;
		facet[addr+2] = v2;
	}
	fclose(msms->ffd); fclose(msms->vfd);
	system("rm tempOut tempIn*");
	//We need to calculate the area of each facet just once, while 
	//the integrand will be calculated for each atom separately.
	float *facetarea = calculate_facetarea(facecount, facet, vertex);
	
	//For each atom, calculate the surface integrals
	//1/alpha_k = (integral/(clk-3)/4pi)^(1/(k-3)), k=4,5,7
	float rev_alpha = calculate_reverse_alpha(facecount, vertexcount, facet,
	 facetarea, vertex, normal, *(m.begin()+atomN));
	
	//Save the data to the atom
	(m.begin()+atomN)->ChunkR6BR(rev_alpha);

        // release  memory
        delete[] vertex;
        delete[] normal;
        delete[] facet;

        delete[] msms;
        delete[] facetarea;

	return 0;
}


////////////////////////////////////////////////////////////////////////////////////

void *MolecularIntegration::open_file_read(const char *filepath, const char *filetype,
    int *natoms) {
  FILE *ffd; // face file
  FILE *vfd; // vertex file
  msms_t *msms;
  char * facefilepath;
  char * vertfilepath;
  char * cp;

  int filenamelen = strlen(filepath);
  facefilepath = (char *) malloc(filenamelen + 10);
  vertfilepath = (char *) malloc(filenamelen + 10);
  strcpy(facefilepath, filepath);
  strcpy(vertfilepath, filepath);
  strcat(facefilepath, ".face");
  strcat(vertfilepath, ".vert");

  //printf("opening face file %s\n", facefilepath);
  //printf("opening vert file %s\n", vertfilepath);

  // require the MSMS output filenames to match what MSMS does
  // If the user selected the .face file or the .vert file, we should
  // be able to cope either way by assigning the right filenames to the
  // right strings and getting the right files opened accordingly.
  
  cp = strstr(facefilepath, ".face");
  if (cp == NULL) {
    cp = strstr(facefilepath, ".vert");
	
//	printf( "   Trying to read file %s\n", facefilepath);
	
    if (cp != NULL) {
       strcpy(cp, ".face");
    } else {
      printf("file names don't match expected MSMS output\n");
      free(facefilepath);
      free(vertfilepath);
      return NULL;
    } 
  }
  cp = strstr(vertfilepath, ".vert");
	
//  printf( "   Trying to read file %s\n", vertfilepath);
	
  if (cp == NULL) {
    cp = strstr(vertfilepath, ".face");
    if (cp != NULL) {
       strcpy(cp, ".vert");
    } else {
      printf("file names don't match expected MSMS output\n");
      free(facefilepath);
      free(vertfilepath);
      return NULL;
    } 
  }
 
  ffd = fopen(facefilepath, "r");
  vfd = fopen(vertfilepath, "r");
  if (!ffd || !vfd) { 
    //printf("%d failed to open either the MSMS face or vertex file\n");
    if (ffd) fclose(ffd);
    if (vfd) fclose(vfd);
    free(facefilepath);
    free(vertfilepath);
    return NULL;
  }
  msms = new msms_t;
  msms->ffd = ffd;
  msms->vfd = vfd;
//  msms->graphics = NULL;
  *natoms = 0;
  return msms;
}

////////////////////////////////////////////////////////////////////////////////////

// Calculating the area of each triangular facet: side a (connecting vertices 0-1)
// times side b (0-2) times sin(angle 1-0-2) / 2. 

float *MolecularIntegration::calculate_facetarea(int facecount, int *facet, float *vertex)
{
  float *facetarea = new float [facecount];

  double x01, y01, z01, x02, y02, z02, a_sq, b_sq, ab, area_sq;	
  int v0, v1, v2, i, addr;
  
  double total_facetarea = 0.;
  
  for (i=0; i<facecount; i++) 
  {
		addr = i * 3;
  		v0 = facet[addr  ];	// indices of vertices making this facet
		v1 = facet[addr+1];
		v2 = facet[addr+2];
		x01 = vertex[3*v1  ] - vertex[3*v0  ];	// edges of the facet
		y01 = vertex[3*v1+1] - vertex[3*v0+1];
		z01 = vertex[3*v1+2] - vertex[3*v0+2];
		x02 = vertex[3*v2  ] - vertex[3*v0  ];
		y02 = vertex[3*v2+1] - vertex[3*v0+1];
		z02 = vertex[3*v2+2] - vertex[3*v0+2];
		a_sq = x01 * x01 + y01 * y01 + z01 * z01;
		b_sq = x02 * x02 + y02 * y02 + z02 * z02;
		ab   = x01 * x02 + y01 * y02 + z01 * z02;
	//	area = a_sq + b_sq - ab;	// this value MUST be positive but let's check
		area_sq = a_sq * b_sq - ab * ab;
	//	if ( area_sq == 0. )
	//	{
	//		continue;
	//   	fprintf(stderr, "WARNING! Face %d (%d,%d,%d) has zero area\n", 
	//			i+1, v0+1, v1+1, v2+1);
	//	}
		if ( area_sq < 0. )
		{
	   		fprintf(stderr, "ERROR! Area of face %d (%d,%d,%d) cannot be calculated (2S^2 = %f)\n", 
				i+1, v0+1, v1+1, v2+1, area_sq);
			area_sq = 0;
		}
		facetarea[i] = sqrt(area_sq)/2;
                //cout << i << "---> " << facetarea[i] << endl;
		total_facetarea += facetarea[i];
  }

  //printf("Total facet area = %.3f A^2, effective radius = %.3f A\n", 
  //	total_facetarea, sqrt( total_facetarea / 4. / M_PI ) );
  
  return facetarea;
}


////////////////////////////////////////////////////////////////////////////////////

// Calculating the area of each triangular facet: side a (connecting vertices 0-1)
// times side b (0-2) times sin(angle 1-0-2) / 2. 

float MolecularIntegration::calculate_totalarea(int facecount, int *facet, float *vertex)
{
  //float *facetarea = new float [facecount];

  double x01, y01, z01, x02, y02, z02, a_sq, b_sq, ab, area_sq;
  int v0, v1, v2, i, addr;

  double total_facetarea = 0.;

  for (i=0; i<facecount; i++)
  {
                addr = i * 3;
        v0 = facet[addr  ];     // indices of vertices making this facet
                v1 = facet[addr+1];
                v2 = facet[addr+2];
                x01 = vertex[3*v1  ] - vertex[3*v0  ];  // edges of the facet
                y01 = vertex[3*v1+1] - vertex[3*v0+1];
                z01 = vertex[3*v1+2] - vertex[3*v0+2];
                x02 = vertex[3*v2  ] - vertex[3*v0  ];
                y02 = vertex[3*v2+1] - vertex[3*v0+1];
                z02 = vertex[3*v2+2] - vertex[3*v0+2];
                a_sq = x01 * x01 + y01 * y01 + z01 * z01;
                b_sq = x02 * x02 + y02 * y02 + z02 * z02;
                ab   = x01 * x02 + y01 * y02 + z01 * z02;
        //      area = a_sq + b_sq - ab;        // this value MUST be positive but let's check
                area_sq = a_sq * b_sq - ab * ab;
        //        if ( area_sq == 0. )
        //        {
        //                continue;
        //      fprintf(stderr, "WARNING! Face %d (%d,%d,%d) has zero area\n", 
        //                      i+1, v0+1, v1+1, v2+1);
        //        }

                if ( area_sq < 0. )
                {
                        fprintf(stderr, "ERROR! Area of face %d (%d,%d,%d) cannot be calculated (2S^2 = %f)\n",
                                i+1, v0+1, v1+1, v2+1, area_sq);
                        area_sq = 0;
                }
                total_facetarea += sqrt(area_sq)/2;    
  }

  //printf("Total facet area = %.3f A^2, effective radius = %.3f A\n", 
  //    total_facetarea, sqrt( total_facetarea / 4. / M_PI ) );

  return total_facetarea;
}





///////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Calculates the surface integral for r^6 (Gricuk) Born radii for one point
// given by its coordinates vec=(x, y, z). 
double MolecularIntegration::calculate_reverse_alpha(int facecount, int vertexcount, int *facet, float *facetarea, 
 float *vertex, float *normal, Atom &a) {
	int i, addr;
	float rx, ry, rz, r2, r4, r2min = 1.e+6;

	// time-saving arrays
	float *r_dot_n = new float[vertexcount];
	float *r6 = new float[vertexcount];
	
	//cout << "cra initial for loop, vertexcount: " << vertexcount << endl;
        float atomr2 = 0;

        bool problems = false;
  
	for (i=0; i<vertexcount; i++) {
		addr = i * 3;
		rx = vertex[addr] - a.X();
		ry = vertex[addr+1] - a.Y();
		rz = vertex[addr+2] - a.Z();

                atomr2 = a.Radius() * a.Radius();
		r_dot_n[i] = rx * normal[addr] + ry * normal[addr+1] + rz * normal[addr+2];

		//if ( r_dot_n[i] <= 0 ) {
		//	cout << i << " ---> " << a.Number() << "   " << vertexcount << endl; 
		//}

		r2 = rx * rx + ry * ry + rz * rz;
                
		//if ( r2 <  0.9*atomr2 )  cout << a.Number() << "  " << i+1 << endl ;
		
		if ( r2 <  0.8*atomr2 ) {
			r6[i] = 0;
			problems = true;
			//cerr << "Distance between atom " << a.Number() << " and vertex " << i+1 << " is too small"<< endl;
		}
		else {
			r4 = r2 * r2;
			r6[i] = r2 * r4;
		}

	}
	
	//cout << "cra going into calculating for loop" << endl;
	
	//Now, we only need to sum up the surface integrals.
	int j, k;
	double sum = 0., weight;
        double partial_sum;

	//cerr << "A: " << endl;
	for (i=0; i<facecount; i++) {
		addr = i * 3;
                
                partial_sum = 0;


		//for (j=0; j<3; j++) {
		k = facet[addr+0];
  		weight = r6[k];
                if (weight <= 0 ) {continue; }
		partial_sum += facetarea[i] * r_dot_n[k] / weight;

                k = facet[addr+1];
                weight = r6[k];
                if (weight <= 0 ) {continue; }
                partial_sum += facetarea[i] * r_dot_n[k] / weight;

                k = facet[addr+2];
                weight = r6[k];
                if (weight <= 0 ) {continue; }
                partial_sum += facetarea[i] * r_dot_n[k] / weight;

  		sum += partial_sum;

		//}

		//cerr << sum << " ";
        	//if (a.Number() == 1 ) cout << endl;
	}

	//cerr << endl << endl;
        //cout << a.Number()<< "--->" <<  sum << endl;	
	//cout << "cra final calculations and returning" << endl;
        

	// sum is the integral; alpha is R (Born radius), 
	// 1/R = [ 1/4pi sum]^{1/3} for n=6 or [ 3/(n-3) * 1/4pi * sum ]^{1/3};	
	// we should also divide sum by 3 because each vertex was counted 3 times; 
	// 4pi comes from physics, (n-3) = 3 from Ostrogradsky theorem		
	sum /= 12. * M_PI;
	
	if (sum <= 0.) {
		sum = 0;
	} 
	else {
      		sum = pow(sum, 0.333333333333); 
	}
    	delete [] r6;
        delete [] r_dot_n;
 	
	if (problems) {
		cerr << "WARNING: Distance between atom: "<< a.Number() << " and some vertices are too small, check your results carefully; you may also try changing the probe radius "<< endl;
		//cout << "BAD " ;
	}
	return sum;
}

////////////////////////////////////////////////////////////////////////////////////

// When directly found Reff <= 0., this procedure calculates an average of the valid
// neighbors' Reff's. 

float MolecularIntegration::averageBornRadius(int bad, int atomcount, float *atomCoord, float *analytRadii)
{
	int i, addr;
	float rij2, weight, average_rev_Reff = 0., norm=0.;

	int addr_bad = 3 * bad;
	for (i=0; i<atomcount; i++)
	{
		if ( i == bad )
			continue;
		addr = 3 * i;
		rij2 = ( atomCoord[addr_bad  ] - atomCoord[addr  ] ) * ( atomCoord[addr_bad  ] - atomCoord[addr  ] ) + 
			   ( atomCoord[addr_bad+1] - atomCoord[addr+1] ) * ( atomCoord[addr_bad+1] - atomCoord[addr+1] ) + 
			   ( atomCoord[addr_bad+2] - atomCoord[addr+2] ) * ( atomCoord[addr_bad+2] - atomCoord[addr+2] );
		if ( rij2 <= NEIGHBORDIST2 && 	// atom atomN is a neighbor to the node
			 analytRadii[i] > 0. )
		{
			weight = 1./rij2;
			average_rev_Reff += 1. / analytRadii[i] * weight;
			norm += weight;
		}
	}
	
	if ( average_rev_Reff > 0. )
		analytRadii[bad] = 1. / (average_rev_Reff / norm);
	else
		fprintf(stderr, "averageBornRadius() failed to produce a positive estimate for analytRadii[%d]=%g\n", 
			bad+1, analytRadii[bad]);

	return analytRadii[bad];
}

////////////////////////////////////////////////////////////////////////////////

int MolecularIntegration::EntireMoleculeR6(Molecule &m, int density, double ProbeRadius, double B )
{
  //cout << "MI start of function" << endl;
	int debug = 0;
	int d_junk;
	const char *c_junk;

	Atoms::iterator itr;
	FILE *temp = fopen("tempOut", "w");
	for (itr = m.begin(); itr != m.end(); itr++) {
		fprintf(temp, "%f %f %f %f\n",
		 itr->X(), itr->Y(), itr->Z(), itr->Radius());
	}
	fclose(temp);
	
	string d;
	stringstream ConvertStream;
	ConvertStream << density;
	ConvertStream >> d;

        //Covert the Probe Radius from double to string 
        std::ostringstream oss;
        oss << ProbeRadius;
        std::string Rw = oss.str();

	//-------------<HACK>----------------//
	string cmd = "msms -if tempOut -prob " + Rw + " -density " + d +
	 " -no_header -of tempIn 1>msms.out 2>/dev/null";
	system(cmd.c_str());
	//-------------</HACK>---------------//

        // Read the SAS Area
        double SASarea; 
        string FirstWord;
        ifstream SASFile;
        SASFile.open("msms.out");
        while (!SASFile.eof()) { 
		SASFile >> FirstWord;
		if (FirstWord == "SAS"  ) {
			SASFile >> SASarea >> SASarea;
			SASFile >> SASarea >> SASarea;
			SASFile >> SASarea >> SASarea;
			SASFile >> SASarea;
			break;
		}
	}
	//cout << SASarea << endl;
	SASFile.close();


	msms_t *msms = (msms_t*) open_file_read("tempIn", c_junk, &d_junk );	
	if (msms == NULL)
		return 1;
	int facecount = 0;
	int vertexcount = 0;
	int *facet;
	float *vertex, *normal;

	int i, j, t, atomN;
	unsigned addr;
  
	float tf;

	//count number of faces
	while (fscanf(msms->ffd, "%d %d %d %d %d", &t, &t, &t, &t, &t) == 5) 
		facecount++;
	rewind(msms->ffd);
	//count number of vertices
	while (fscanf(msms->vfd, "%f %f %f %f %f %f %d %d %d", 
         &tf, &tf, &tf, &tf, &tf, &tf, &t, &t, &t) == 9)
		vertexcount++;
	rewind(msms->vfd);
	// simple sanity check to insure we have at least one usable triangle
	if (facecount < 1 || vertexcount < 3) {
   		fprintf(stderr,"Insufficient data facecount=%d, vertexcount=%d", 
		 facecount, vertexcount);
		return 1;
	}
	
	// allocate storage for vertex and normal data
	vertex = new float[3 * vertexcount];
	normal = new float[3 * vertexcount];
	facet  = new int  [3 * facecount  ];
	
  
	// read in the vertex data
	for (i=0; i<vertexcount; i++) {
		addr = i * 3;
		//Not used
		int atomid, l0fa, l;
		fscanf(msms->vfd, "%f %f %f %f %f %f %d %d %d",
           	 &vertex[addr], &vertex[addr+1], &vertex[addr+2], 
           	 &normal[addr], &normal[addr+1], &normal[addr+2], &l0fa, &atomid, &l);
  	}
	// read in the facet data
	for (i=0; i<facecount; i++) {
		int v0, v1, v2, surftype, ana;
		// read in the next facet
		fscanf(msms->ffd, "%d %d %d %d %d", &v0, &v1, &v2, &surftype, &ana);
		v0--; // convert from 1-based indexing to 0-based indexing
		v1--;
		v2--;
		addr = i * 3;
		facet[addr] = v0;
		facet[addr+1] = v1;
		facet[addr+2] = v2;
	}
	fclose(msms->ffd); fclose(msms->vfd);
	
	//We need to calculate the area of each facet just once, while 
 	//the integrand will be calculated for each atom separately.
	float *facetarea = calculate_facetarea(facecount, facet, vertex);
	// For each atom, calculate the surface integrals
  
	// 1/alpha_k = (integral/(clk-3)/4pi)^(1/(k-3)), k=4,5,7
	float *reverse_alpha = new float[m.size()];
	float *analytRadii = new float[m.size()];


	for (i = 0; i < m.size(); i++) {
		addr = atomN * 3;
		reverse_alpha[i] = calculate_reverse_alpha(facecount,
		 vertexcount, facet, facetarea, vertex, normal, *(m.begin()+i));
		(m.begin()+i)->MolecularR6Sum(reverse_alpha[i] + B);
	}

        // Realease memory
        delete[] vertex;
        delete[] normal;
        delete[] facet;

        delete[] msms;
        delete[] facetarea;

        delete[] reverse_alpha;
        delete[] analytRadii;
	return 0;
}

////////////////////////////////////////////////////////////////////////////////

int MolecularIntegration::NumericalNeckIntegration(Molecule &m, int density, double ProbeRadius)
{
  //cout << "MI start of function" << endl;
	int debug = 0;
	int d_junk;
	const char *c_junk;

	Atoms::iterator itr;
	FILE *temp = fopen("tempOut", "w");
	for (itr = m.begin(); itr != m.end(); itr++) {
		fprintf(temp, "%f %f %f %f\n",
		 itr->X(), itr->Y(), itr->Z(), itr->Radius());
	}
	fclose(temp);
	
	string d;
	stringstream ConvertStream;
	ConvertStream << density;
	ConvertStream >> d;

        //Covert the Probe Radius from double to string 
	std::ostringstream oss;
 	oss << ProbeRadius;
	std::string Rw = oss.str();


	//-------------<HACK>----------------//
	string cmd = "msms -if tempOut -prob " + Rw + " -density " + d +
	 " -no_header -of tempIn 1>/dev/null 2>/dev/null";
	system(cmd.c_str());
	//-------------</HACK>---------------//
	
	msms_t *msms = (msms_t*) open_file_read("tempIn", c_junk, &d_junk );	
	if (msms == NULL)
		return 1;
	int facecount = 0;
	int vertexcount = 0;
	int *facet;
	float *vertex, *normal;

	int i, j, t, atomN;
	unsigned addr;
  
	float tf;

	//count number of faces
	while (fscanf(msms->ffd, "%d %d %d %d %d", &t, &t, &t, &t, &t) == 5) 
		facecount++;
	rewind(msms->ffd);
	//count number of vertices
	while (fscanf(msms->vfd, "%f %f %f %f %f %f %d %d %d", 
         &tf, &tf, &tf, &tf, &tf, &tf, &t, &t, &t) == 9)
		vertexcount++;
	rewind(msms->vfd);
	// simple sanity check to insure we have at least one usable triangle
	if (facecount < 1 || vertexcount < 3) {
   		fprintf(stderr,"Insufficient data facecount=%d, vertexcount=%d", 
		 facecount, vertexcount);
		return 1;
	}
	
	// allocate storage for vertex and normal data
	vertex = new float[3 * vertexcount];
	normal = new float[3 * vertexcount];
	facet  = new int  [3 * facecount  ];
	
  
	// read in the vertex data
	for (i=0; i<vertexcount; i++) {
		addr = i * 3;
		//Not used
		int atomid, l0fa, l;
		fscanf(msms->vfd, "%f %f %f %f %f %f %d %d %d",
           	 &vertex[addr], &vertex[addr+1], &vertex[addr+2], 
           	 &normal[addr], &normal[addr+1], &normal[addr+2], &l0fa, &atomid, &l);
  	}
	// read in the facet data
	for (i=0; i<facecount; i++) {
		int v0, v1, v2, surftype, ana;
		// read in the next facet
		fscanf(msms->ffd, "%d %d %d %d %d", &v0, &v1, &v2, &surftype, &ana);
		v0--; // convert from 1-based indexing to 0-based indexing
		v1--;
		v2--;
		addr = i * 3;
		facet[addr] = v0;
		facet[addr+1] = v1;
		facet[addr+2] = v2;
	}
	fclose(msms->ffd); fclose(msms->vfd);
	
	//We need to calculate the area of each facet just once, while 
 	//the integrand will be calculated for each atom separately.
	float *facetarea = calculate_facetarea(facecount, facet, vertex);
	// For each atom, calculate the surface integrals
  
	// 1/alpha_k = (integral/(clk-3)/4pi)^(1/(k-3)), k=4,5,7
	float *reverse_alpha = new float[m.size()];
	float *analytRadii = new float[m.size()];


	for (i = 0; i < m.size(); i++) {
		addr = atomN * 3;
		reverse_alpha[i] = calculate_reverse_alpha(facecount,
		 vertexcount, facet, facetarea, vertex, normal, *(m.begin()+i));
		(m.begin()+i)->MolecularR6Sum(reverse_alpha[i]);
	}

        // Realease memory
        delete[] vertex;
        delete[] normal;
        delete[] facet;

        delete[] msms;
        delete[] facetarea;

        delete[] reverse_alpha;
        delete[] analytRadii;

	return 0;
}

int MolecularIntegration::ComputeAverageRadius(Molecule &m, int density, double ProbeRadius)
{
  //cout << "MI start of function" << endl;
	int debug = 0;
	int d_junk;
	const char *c_junk;

	// If just one atom the no need all this calculation
	if (m.size() == 1) {
     		m.Xav( m.begin()->X()); /* now it's the center */
        	m.Yav( m.begin()->Y());
        	m.Zav( m.begin()->Z());
		m.Rav( m.begin()->OrigRadius() );
                //cout << "Atom .....  " << m.begin()->Type()  << endl;
	        return 0;
	}

	

	Atoms::iterator itr;
	FILE *temp = fopen("tempOut", "w");
	for (itr = m.begin(); itr != m.end(); itr++) {
		fprintf(temp, "%f %f %f %f\n",
		 itr->X(), itr->Y(), itr->Z(), itr->Radius());
	}
	fclose(temp);
	
	string d;
	stringstream ConvertStream;
	ConvertStream << density;
	ConvertStream >> d;

        //Covert the Probe Radius from double to string 
	std::ostringstream oss;
 	oss << ProbeRadius;
	std::string Rw = oss.str();

	//-------------<HACK>----------------//
	string cmd = "msms -if tempOut -prob " + Rw + " -density " + d +
	 " -no_header -of tempIn 1>/dev/null 2>/dev/null";
	system(cmd.c_str());
	//-------------</HACK>---------------//
	
	msms_t *msms = (msms_t*) open_file_read("tempIn", c_junk, &d_junk );	
	if (msms == NULL)
		return 1;
	int facecount = 0;
	int vertexcount = 0;
	int *facet;
	float *vertex, *normal;

	int i, j, t, atomN;
	unsigned addr;
  
	float tf;

	//count number of faces
	while (fscanf(msms->ffd, "%d %d %d %d %d", &t, &t, &t, &t, &t) == 5) 
		facecount++;
	rewind(msms->ffd);
	//count number of vertices
	while (fscanf(msms->vfd, "%f %f %f %f %f %f %d %d %d", 
         &tf, &tf, &tf, &tf, &tf, &tf, &t, &t, &t) == 9)
		vertexcount++;
	rewind(msms->vfd);
	// simple sanity check to insure we have at least one usable triangle
	if (facecount < 1 || vertexcount < 3) {
   		fprintf(stderr,"Insufficient data facecount=%d, vertexcount=%d", 
		 facecount, vertexcount);
		return 1;
	}
	
	// allocate storage for vertex and normal data
	vertex = new float[3 * vertexcount];
	normal = new float[3 * vertexcount];
	facet  = new int  [3 * facecount  ];
	
  
	// read in the vertex data
	for (i=0; i<vertexcount; i++) {
		addr = i * 3;
		//Not used
		int atomid, l0fa, l;
		fscanf(msms->vfd, "%f %f %f %f %f %f %d %d %d",
           	 &vertex[addr], &vertex[addr+1], &vertex[addr+2], 
           	 &normal[addr], &normal[addr+1], &normal[addr+2], &l0fa, &atomid, &l);
  	}
	// read in the facet data
	for (i=0; i<facecount; i++) {
		int v0, v1, v2, surftype, ana;
		// read in the next facet
		fscanf(msms->ffd, "%d %d %d %d %d", &v0, &v1, &v2, &surftype, &ana);
		v0--; // convert from 1-based indexing to 0-based indexing
		v1--;
		v2--;
		addr = i * 3;
		facet[addr] = v0;
		facet[addr+1] = v1;
		facet[addr+2] = v2;
	}
	fclose(msms->ffd); fclose(msms->vfd);
	
	//We need to calculate the area of each facet just once, while 
 	//the integrand will be calculated for each atom separately.
	double total_area = calculate_totalarea(facecount, facet, vertex);
	
        
	//calculate the radius of the molecule
	double Radius = sqrt(total_area/(4.0 * M_PI)); 

	// Rounding to the third decimal
	Radius = ( int ((Radius + 0.0005)*1000)) / 1000.0;
	m.Rav(Radius); 
	
	// Claculate the center of mass of the molecule (mass = r^3)
	double Xav = 0; // X coordinate of the center og mass
	double Yav = 0; // Y ....
	double Zav = 0; // Z ....

	double molecule_mass = 0;

	for (i = 0; i < m.size(); i++) {

		double x_i = (m.begin() + i)->X();
		double y_i = (m.begin() + i)->Y();
		double z_i = (m.begin() + i)->Z();

		double r3_i = (m.begin() + i)->OrigRadius();
                r3_i = r3_i*r3_i*r3_i;

                /* Calculating molecule mass and center of mass */
                Xav += x_i * r3_i; /* atom's "mass" is simply r^3 */
                Yav += y_i * r3_i;
                Zav += z_i * r3_i;
                molecule_mass += r3_i;

        }
        m.Xav(Xav / molecule_mass); /* now it's the center */
        m.Yav(Yav / molecule_mass);
        m.Zav(Zav / molecule_mass);

        // Realease memory
        delete[] vertex;
        delete[] normal;
        delete[] facet;

        delete[] msms;

        return 0;
}
               
