/////////////////////////////////////////////////////////////
//Programer      : Richard 
//Modifiev by    : Boris Aguilar 
//Date of Modification : 8/4/2009
//
//Program NAme : NeckParams
//OS           : Linux, Fedora and ubuntu
//compiler     : g++
//
//Usage        : This program is called by bornradius.cpp to get 
//               the calues of the neck table written in the file
//               It is aslo called by the function Analytic::calculate
//	         to get the paramteres Aij and Bij.
//
//Description  :  
//  
     

#include "NeckParams.h"

#define TBL_SIZE 128
const double eps = 0.01;

//////////////////////////////////////////////
// CONSTRUCTOR
//
// Purpose  :  It will initialize a table of 128 Nodes 
//             The probe radius for Necks will be 1.4 if no
//             probe radius value is given 
// 
NeckParams::NeckParams() {
  
  r_w = 1.4; // Water Radius will be 1,4 for default -B.A
 
  dirty = false; // This will indicate if we need to update the neck file 

  table = new Node[TBL_SIZE];
  int i;
  for (i = 0; i < TBL_SIZE; i++) {
    table[i].Ri = 0;
    table[i].Rj = 0;
    table[i].next = NULL;
  }
}
////////////////////////////////////////////////////////////
// Purpose  :  load the table with values of Ri, Rj, Aij and Bij
//               with the values in the NEck File if exist 
//
// Parameters : 
//   f : Name of the Neck file. If no specific name is specified
//       then "temp.neck" is used
//
// Called  by : main() in bornRadius.cpp

void NeckParams::loadFromFile(string f) {
  fileName = f;
  ifstream in(fileName.c_str());
  int count = 0;
  while (in && !in.eof()) {
    double Ri, Rj, Aij, Bij;
    in >> Ri >> Rj >> Aij >> Bij;  
    if (!in) break;
    
    Node* n = &table[hash(Ri, Rj)];
    while (n->next != NULL) n = n->next;
    if (!Functions::equal(n->Ri,0,eps) || !Functions::equal(n->Rj,0,eps)) {
      n->next = new Node;
      n = n->next;
    }
    n->Ri = Ri; n->Rj = Rj; n->Aij = Aij; n->Bij = Bij;  n->next = NULL;
    count++;
  }
  //cout << "Loaded " << count << " neck data from file." << endl;
}

/////////////////////////////////////////////////////////////
//
//Parameters:
//  Ri : Original radius of atom i (atom which born radius we are calculating
//  Rj : Orignial Radius of Atom j (atom close to amot i )
//
//Returns :   Index for the table
//
//called by : loadFromFile() and FindNode()

int NeckParams::hash(double Ri, double Rj) {
  return (((int)(Ri*100))*100 + ((int)(Rj*100))) % TBL_SIZE;
}

/////////////////////////////////////////////////////
// purpose : 
//
// Parameters:
//   Ri : Radius of atom i (in which we are computing the born radius)
//   Rj : Radius of atom j (close to i)
//
// Output : Pointer to a structure Node (which ciontains Aij and Bij)
//
// Called by "Neck()" in Analytic.cpp
Node* NeckParams::lookup(double Ri, double Rj) {
  
  Node *n = findNode(hash(Ri, Rj), Ri, Rj);
  
  if (!Functions::equal(Ri, n->Ri,eps) || !Functions::equal(Rj, n->Rj,eps)) {
    //cout << "Unknown radii " << Ri << " and " << Rj << "...computing neck." << endl;
    findNeckParam3(Ri, Rj, n); // fill Aij and Bij of node n
  }
  
  return n;
}

Node* NeckParams::findNode(int index, double Ri, double Rj) {
  //cout << "Starting findNode with index " << index << endl;
  Node *n = &table[index];
  //cout << "Past getting n in findNode" << endl;
  while (true) {
    if (Functions::equal(Ri, n->Ri,eps) && Functions::equal(Rj, n->Rj,eps)) return n;
    if (n->next == NULL) break;
    //cout << "Didn't break in loop in findNode." << endl;
    n = n->next;
  }
  //Got to the end of the list, couldn't find it.  Allocate a new node.
  if (Functions::equal(n->Ri, 0,eps) && Functions::equal(n->Rj, 0,eps)) return n;
  n = n->next = new Node;
  n->Ri = 0; n->Rj = 0; n->Aij = 0; n->Bij = 0; n->next = NULL;
  return n;
}

NeckParams::~NeckParams() {
  ofstream o;
  if (fileName != "" && dirty) {
    o.open(fileName.c_str());
  }
  int i;
  for (i = 0; i < TBL_SIZE; i++) {
    Node *n = table + i;

    //---------------- Modified By Boris Aguilar ---------------------------------
    // without this code program generates memory allocation problems
    if (n != NULL && (!Functions::equal(0, n->Ri,eps) || !Functions::equal(0, n->Rj,eps))) {
      if (o)  o << n->Ri << "\t" << n->Rj << "\t" << n->Aij << "\t"<< n->Bij << endl;
    }
    n = n->next;
    //--------------------------------------------------------------------------------
    while (n != NULL && (!Functions::equal(0, n->Ri,eps) || !Functions::equal(0, n->Rj,eps))) {
      Node *next = n->next;
      if (o)  o << n->Ri << "\t" << n->Rj << "\t" << n->Aij << "\t" <<n->Bij  << endl;
      delete n;
      n = next;
    }
  }
  delete []table;
}



void NeckParams::findNeckParam(double Ri, double Rj, Node* n) {

  dirty = true;
  double Rw = WaterRadius();
  double start = Ri, end = Ri+Rj+2*Rw, step = 0.01, d;
  double max = 0, maxValue = 0;
  
  //cout << "Searching for best fit of the neck:" << endl;

  for (d = start; d <= end; d += step) {
    Atom ai(0,0,0,Ri,1);
    Atom aj(d,0,0,Rj,2);
    Molecule m; m.addAtom(ai); m.addAtom(aj);
    MolecularIntegration::NumericalNeckIntegration(m, 40, Rw);
    
    Atoms::iterator itr = m.begin();

    double R = itr->MolecularR6Sum();
    double integral = (4.*M_PI)/3. * (1/(Ri*Ri*Ri) - R*R*R);
    double term = Rj / (d*d - Rj*Rj);
    double gry;
    if (d > Ri+Rj) {
      gry = ( (4.*M_PI/3.) ) * term*term*term;
    }
    else {
      double v0 = 16 * d;
      double v1 = d + 3*Rj;
      double v2 = d + Rj;
      double v3 = 3*(Rj*Rj - Ri*Ri - (d-Ri)*(d-Ri))+2*d*Ri;
      double v4 = Ri*Ri*Ri*Ri;
      gry = ( (4.*M_PI)/3. ) * (1/v0) * (v1/(v2*v2*v2) + v3/v4);
    }

   // find the maximun value  
    if (integral - gry > maxValue) {
      max = d;
      maxValue = integral - gry;
    }
  }

  /* Modified by Boris Aguilar */  
  //---------------------------
  // find the two parameteres needed to fit the neck integral 
  // to our proposed function
  double Bij = 2*max - Ri -Rj - 2*Rw;
  double Aij = maxValue / (pow(max -Bij ,4)*pow(Ri+Rj+2*Rw - max,4));
  /*-------------------------------------------------------------*/

  n->Ri = Ri;
  n->Rj = Rj;
  n->Aij = Aij;
  n->Bij = Bij;
}

// Acces the Water Radius used only for Necks -B.A
double NeckParams::WaterRadius() const {
  return r_w;
}

// Modify The Water Radius used inly for Necks -B.A
void NeckParams::WaterRadius(double paramRw) {
  r_w = paramRw;
}




//
//
//
//
//


void NeckParams::findNeckParam2(double Ri, double Rj, Node* n) {

  dirty = true;
  double Rw = WaterRadius();
  double start = Ri, end = Ri+Rj+2*Rw, step = 0.01, d;
  double max = 0, maxValue = 0;

  int p = 4;
  int q = 3;

  //  if d < start the neck is zero
  start = Ri-Rj;
  if (start < 0) start = -1*start; // absolute value
  

  double Bij = start;
  //cout << "Searching for best fit of the neck:" << endl;


  //------------------------------------------------
  // find the maximun value of the proposed function,
  //  and the distance at which it happens
  for (d = start; d <= end; d += step) {

    double neck = pow(d -Bij ,p)*pow(Ri+Rj+2*Rw - d,q);
    
    // find the maximun value  
    if ( neck  > maxValue) {
      max = d;
      maxValue = neck;
    }
  } 
  d = max;
  //--------------------------------------------------

  // ----------------------------------------------------------------------
  // Compute the real value of the neck at d = max
  // ( integral - gry)
  //-----------------------------------------------------------------------
  Atom ai(0,0,0,Ri,1);
  Atom aj(d,0,0,Rj,2);
  Molecule m; m.addAtom(ai); m.addAtom(aj);
  MolecularIntegration::NumericalNeckIntegration(m, 40, 1.4);
    
  Atoms::iterator itr = m.begin();
  double R = itr->MolecularR6Sum();
  double integral = (4.*M_PI)/3. * (1/(Ri*Ri*Ri) - R*R*R);
  double term = Rj / (d*d - Rj*Rj);
  double gry;
  if (d > Ri+Rj) {
      gry = ( (4.*M_PI/3.) ) * term*term*term;
  }
  else {
      double v0 = 16 * d;
      double v1 = d + 3*Rj;
      double v2 = d + Rj;
      double v3 = 3*(Rj*Rj - Ri*Ri - (d-Ri)*(d-Ri))+2*d*Ri;
      double v4 = Ri*Ri*Ri*Ri;
      gry = ( (4.*M_PI)/3. ) * (1/v0) * (v1/(v2*v2*v2) + v3/v4);
  }
  //---------------------------------------------------------------------

  //---------------------------
  // find the two parameteres needed to fit the neck integral 
  // to our proposed function
  //double Bij = 2*max - Ri -Rj - 2*Rw;
  double Aij = (integral - gry) / (pow(max -Bij ,p)*pow(Ri+Rj+2*Rw - max,q));
  /*-------------------------------------------------------------*/

  n->Ri = Ri;
  n->Rj = Rj;
  n->Aij = Aij;
  n->Bij = Bij;
}



//------------------------------------------------------------
// This function will work for many probe radius,
// Developed by Boris Aguilar

void NeckParams::findNeckParam3(double Ri, double Rj, Node* n) {

  dirty = true;
  double Rw = WaterRadius();
  double start = Ri, end = Ri+Rj+2*1.4, step = 0.01, d;
  double max = 0, maxValue = 0;

  int p = 4;
  int q = 4;
  
  //cout << "Searching for best fit of the neck:" << endl;

  for (d = start; d <= end; d += step) {
    Atom ai(0,0,0,Ri,1);
    Atom aj(d,0,0,Rj,2);
    Molecule m; m.addAtom(ai); m.addAtom(aj);
    MolecularIntegration::NumericalNeckIntegration(m, 40, 1.4);
    
    Atoms::iterator itr = m.begin();

    double R = itr->MolecularR6Sum();
    double integral = (4.*M_PI)/3. * (1/(Ri*Ri*Ri) - R*R*R);
    double term = Rj / (d*d - Rj*Rj);
    double gry;
    if (d > Ri+Rj) {
      gry = ( (4.*M_PI/3.) ) * term*term*term;
    }
    else {
      double v0 = 16 * d;
      double v1 = d + 3*Rj;
      double v2 = d + Rj;
      double v3 = 3*(Rj*Rj - Ri*Ri - (d-Ri)*(d-Ri))+2*d*Ri;
      double v4 = Ri*Ri*Ri*Ri;
      gry = ( (4.*M_PI)/3. ) * (1/v0) * (v1/(v2*v2*v2) + v3/v4);
    }

   // find the maximun value  
    if (integral - gry > maxValue) {
      max = d;
      maxValue = integral - gry;
    }
  }

  /* Modified by Boris Aguilar */  
  //---------------------------
  // find the two parameteres needed to fit the neck integral 
  // to our proposed function
  double Bij = 2*max - Ri -Rj - 2*1.4;



  //------------------------------------------------
  // find the maximun value of the proposed function,
  //  and the distance at which it happens
  maxValue = 0;
  for (d = Bij; d <= (Ri+Rj+2*Rw) ; d += 0.01) {

    double neck = pow(d -Bij ,p)*pow(Ri+Rj+2*Rw - d,q);

    // find the maximun value  
    if ( neck  > maxValue) {
      max = d;
      maxValue = neck;
    }
  }
  d = max;
  //--------------------------------------------------


  // ----------------------------------------------------------------------
  // Compute the real value of the neck at d = max
  // ( integral - gry)
  //-----------------------------------------------------------------------
  Atom ai(0,0,0,Ri,1);
  Atom aj(d,0,0,Rj,2);
  Molecule m; m.addAtom(ai); m.addAtom(aj);
  MolecularIntegration::NumericalNeckIntegration(m, 40, 1.4);

  Atoms::iterator itr = m.begin();
  double R = itr->MolecularR6Sum();
  double integral = (4.*M_PI)/3. * (1/(Ri*Ri*Ri) - R*R*R);
  double term = Rj / (d*d - Rj*Rj);
  double gry;
  if (d > Ri+Rj) {
      gry = ( (4.*M_PI/3.) ) * term*term*term;
  }
  else {
      double v0 = 16 * d;
      double v1 = d + 3*Rj;
      double v2 = d + Rj;
      double v3 = 3*(Rj*Rj - Ri*Ri - (d-Ri)*(d-Ri))+2*d*Ri;
      double v4 = Ri*Ri*Ri*Ri;
      gry = ( (4.*M_PI)/3. ) * (1/v0) * (v1/(v2*v2*v2) + v3/v4);
  }
  //---------------------------------------------------------------------

  //---------------------------
  // find the two parameteres needed to fit the neck integral 
  // to our proposed function
  //double Bij = 2*max - Ri -Rj - 2*Rw;
  double Aij = (integral - gry) / (pow(d -Bij ,p)*pow(Ri+Rj+2*Rw - d,q));
  /*-------------------------------------------------------------*/

  n->Ri = Ri;
  n->Rj = Rj;
  n->Aij = Aij;
  n->Bij = Bij;
}
