#include "VDWIntegration.h"

void VDWIntegration::calcIntegralR4(Molecule &m, int depth) {
	cout << "*****Riemann Sums for R4 Integral*****" << endl;
	cout << "*****    Depth: " << depth << "    *****" << endl;
	Atoms::iterator i, j;
	m.initalizeBounds();
	for (i = m.begin(); i != m.end(); i++) {
		double sum = 0;
		double x, y, z;
		double initalStepSize = 2;
		for (x = m.XMin(); x < m.XMax(); x += initalStepSize) {
			for (y = m.YMin(); y < m.YMax(); y += initalStepSize) {
				for (z = m.ZMin(); z < m.ZMax(); z += initalStepSize) {
					sum += integrateBlockR4(m, x, y, z, initalStepSize, 0, depth, *i);
				}
			}
		}
		double beta = 0;
		double result = 1 / (i->Radius() - beta) - 0.25/M_PI * sum;
		i->VdwR4Sum(sum);
		cout << "Atom " << i->Number() << " Sum: " << sum << " Result: " << result << endl;
	}
	cout << endl;
}

double VDWIntegration::integrateBlockR4(Molecule &m, double xMin, double yMin, double zMin,
 double size, int step, int maxSteps, Atom &a_i) {
	Atoms::iterator j;
	double sum = 0;
 	if (step == maxSteps) {
 		double x = xMin + size/2;
 		double y = yMin + size/2;
 		double z = zMin + size/2;
 		if (a_i.distance(x,y,z) < a_i.Radius()) return 0;
 		for (j = m.begin(); j != m.end(); j++) {
			if (*j == a_i) continue;
 			if (j->distance(x, y, z) < j->Radius()) {
				double d = a_i.distance(x, y, z);
				sum += size*size*size * 1/(d*d*d*d);
				break;
			}
		}
		return sum;
	}
	
	//0.866025404 = sqrt(3)/2
	double radius = 0.866025404 * size;
	for (j = m.begin(); j != m.end(); j++) {
		if (*j == a_i) continue;
		if (j->distance(xMin+size/2, yMin+size/2, zMin+size/2) <= j->Radius()+radius) {
			break;
		}
	}
	if (j != m.end()) {
		double newSize = size / 2;
		double x, y, z;
		for (x = xMin; x < xMin+size; x+=newSize) {
			for (y = yMin; y < yMin+size; y+=newSize) {
				for (z = zMin; z < zMin+size; z+=newSize) {
					sum += integrateBlockR4(m, x, y, z, newSize, step+1, maxSteps, a_i);
				}
			}
		}
	}
	return sum;
}

void VDWIntegration::calcIntegralR6(Molecule &m, int depth) {
	//cout << "*****Riemann Sums for R6 Integral*****" << endl;
	//cout << "*****    Depth: " << depth << "    *****" << endl;
	Atoms::iterator i, j;
	m.initalizeBounds();
	for (i = m.begin(); i != m.end(); i++) {
		double sum = 0;
		double x, y, z;
		double initalStepSize = 2;
		for (x = m.XMin(); x < m.XMax(); x += initalStepSize) {
			for (y = m.YMin(); y < m.YMax(); y += initalStepSize) {
				for (z = m.ZMin(); z < m.ZMax(); z += initalStepSize) {
					sum += integrateBlockR6(m, x, y, z, initalStepSize, 0, depth, *i);
				}
			}
		}
		sum *= 3/(4.0*M_PI);
		double b_i = i->Radius();
		double result = Functions::CubeRoot(1/(b_i*b_i*b_i) - sum);
		i->VdwR6Sum(sum);
		cout << result << endl;
	}
	cout << endl;
}

double VDWIntegration::integrateBlockR6(Molecule &m, double xMin, double yMin, double zMin,
 double size, int step, int maxSteps, Atom &a_i) {
	Atoms::iterator j;
	double sum = 0;
 	if (maxSteps - a_i.distance(xMin+size/2,yMin+size/2,zMin+size/2)/3 < step) {
 		//cout << "Calculating at step " << step << endl;
 		double x = xMin + size/2;
 		double y = yMin + size/2;
 		double z = zMin + size/2;
 		if (a_i.distance(x,y,z) < a_i.Radius()) return 0;
 		for (j = m.begin(); j != m.end(); j++) {
			if (*j == a_i) continue;
 			if (j->distance(x, y, z) < j->Radius()) {
				double d = a_i.distance(x, y, z);
				sum += size*size*size * 1/(d*d*d*d*d*d);
				break;
			}
		}
		return sum;
	}
	
	//0.866025404 = sqrt(3)/2
	double radius = 0.866025405 * size;
	for (j = m.begin(); j != m.end(); j++) {
		if (*j == a_i) continue;
		if (j->distance(xMin+size/2, yMin+size/2, zMin+size/2) <= j->Radius()+radius) {
			break;
		}
	}
	if (j != m.end()) {
		double newSize = size / 2;
		double x, y, z;
		for (x = xMin; x < xMin+size; x+=newSize) {
			for (y = yMin; y < yMin+size; y+=newSize) {
				for (z = zMin; z < zMin+size; z+=newSize) {
					sum += integrateBlockR6(m, x, y, z, newSize, step+1, maxSteps, a_i);
				}
			}
		}
	}
	return sum;
}

void VDWIntegration::calcIntegralR6Spherical(Molecule &m, double stepSize) {
	cout << "*****Spherical coords for R6 Integral*****" << endl;
	cout << "*****    Stepsize: " << stepSize << "    *****" << endl;
	Atoms::iterator i, j;
	for (i = m.begin(); i != m.end(); i++) {
		double minR = i->Radius();
		double maxR = m.begin()->distance(*i) + m.begin()->Radius();
		double minP = 0;
		double maxP = M_PI;
		double minT = 0;
		double maxT = 2*M_PI;
		for (j = m.begin()+1; j != m.end(); j++) {
			if (j->distance(*i) + j->Radius() > maxR) {
				maxR = j->distance(*i) + j->Radius();
			}
		}
		
		double r, phi, theta;
		double sum = 0;
		for (r = minR; r < maxR; r += stepSize) {
			for (phi = minP; phi < maxP; phi += stepSize) {
				for (theta = minT; theta < maxT; theta += stepSize) {
					double midR = r + stepSize/2;
					double midP = phi + stepSize/2;
					double midT = theta + stepSize/2;
					
					double x = midR * sin(midP) * cos(midT) + i->X();
					double y = midR * sin(midP) * sin(midT) + i->Y();
					double z = midR * cos(midP) + i->Z();

					double d = midR;
					
					if (d >= i->Radius()) {
						for (j = m.begin(); j != m.end(); j++) {
							if (*i == *j) continue; 
							if (j->distance(x, y, z) < j->Radius()) {
								sum += ( (r+stepSize)*(r+stepSize)*(r+stepSize) - r*r*r ) / 3 *
									 ( cos(phi) - cos(phi + stepSize) ) * (stepSize) / (d*d*d*d*d*d);
								break;
							}
						}
					}
				}
			}
		}
		cout << i->Number() << "\t" << 3/(4.0*M_PI)*sum << endl;
	}
}

double VDWIntegration::volume(Molecule &m, int depth) {
	m.initalizeBounds();
	double initalStepSize = 1;
	double sum = 0;
	double x, y, z;
	for (x = m.XMin(); x < m.XMax(); x += initalStepSize) {
		for (y = m.YMin(); y < m.YMax(); y += initalStepSize) {
			for (z = m.ZMin(); z < m.ZMax(); z += initalStepSize) {
				sum += integrateBlockVolume(m, x, y, z, initalStepSize, 0, depth);
			}
		}
	}
	return sum;
}

double VDWIntegration::integrateBlockVolume(Molecule &m, double xMin, double yMin, double zMin,
 double size, int step, int maxSteps) {
 
	Atoms::iterator j;
	double sum = 0;
 	if (step == maxSteps) {
 		//cout << "Calculating at step " << step << endl;
 		double x = xMin + size/2;
 		double y = yMin + size/2;
 		double z = zMin + size/2;
 		for (j = m.begin(); j != m.end(); j++) {
 			if (j->distance(x, y, z) < j->Radius()) {
				return size*size*size;
			}
		}
		return 0;
	}
	
	//0.866025404 = sqrt(3)/2
	double radius = 0.866025405 * size;
	for (j = m.begin(); j != m.end(); j++) {
		if (j->distance(xMin+size/2, yMin+size/2, zMin+size/2) <= j->Radius()+radius) {
			break;
		}
	}
	if (j != m.end()) {
		double newSize = size / 2;
		double x, y, z;
		for (x = xMin; x < xMin+size; x+=newSize) {
			for (y = yMin; y < yMin+size; y+=newSize) {
				for (z = zMin; z < zMin+size; z+=newSize) {
					sum += integrateBlockVolume(m, x, y, z, newSize, step+1, maxSteps);
				}
			}
		}
	}
	return sum;
}
