#include <iostream>
#include <iomanip>
#include <string>
#include <vector>
#include <ctime>
#include <algorithm>

#include <sys/time.h>

using namespace std;

#ifdef __GNUC__
static void debug(const char*, ...) __attribute__ ((format(printf, 1, 2)));
static void fatal(const char*, ...) __attribute__ ((format(printf, 1, 2)))
                                    __attribute__ ((__noreturn__));
#else
static void debug(const char*, ...);
static void fatal(const char*, ...);
#endif /* __GNUC__ */

#include "Atom.h"
#include "Molecule.h"
#include "Functions.h"
#include "VDWIntegration.h"
#include "MolecularIntegration.h"
#include "Analytic.h"
#include "NeckParams.h"
#include "GBorn.h"
//Things in bornRadius.h may depend on includes above.  It MUST appear last
#include "bornRadius.h"

#include "elsize.h"

void dispAnalytic(Molecule &m, bool overlaps, bool necks, bool thirdParty);
void dispAnalyticBornRadius(Molecule &m);
bool dispMolecularIntegrationR6(Molecule &m, string rinvFile );

//void dispAnalyticBornRadius(Molecule &m , double shift);

int main(int argc, char * argv[]) {

   //Parameters from the command line
   string pqrFile="", psfFile="", prmtopFile=""; 
   string chunkFile="temp.chunk", neckFile="", rinvFile="";
   string etype = "";
   string DGijFile = "";
   string rinvIn= "";
   //string ParameterFile = "LennardJones.dat";
   string ParameterFile = "";
   string ParameterFile_in = "";
   string inradii = "";

   bool XYZformat=false, vanDerWaalsVolume=false, analytic=false,
        overlapTest=false, molecularIntegrationR6=true, foo=false,
  	findNecks=false, bar=false, vdwIntegration=false;
   bool nolabel=false; 
   bool chagb_flag = false;
   bool gb_flag=false;
   bool gbc = false;
   bool zap9 = false; // flag to set zap9 radii (topology needed)
   bool NeckNumbers = false;
   bool solv_only = false;
   bool elsize_input = false;
	
   int depthR6, density=6, chunkDepth=3, numOverlaps=0;

   double A=1., alpha=1., beta=18.4377, gamma=313.1717, 
          delta=0.6211, omega=0.4058, B=0.028, Neck_Rw=1.40;
   double B_in = -1.0;	
   double ProbeRadius=1.4; 
   bool ProbeRadius_in=false; 

   double alpha_alpb=0.57;
   double ROH=0.58588;
   NeckParams neckParams;

   Atoms::iterator j;
   
   // Parameters for GB
   double diel_ext = 80;
   double diel_int = 1;
   double kappa = 0;
   
   // Surface tension
   double gammaSASA=0.005;    //default value 
   double gammaSASA_in = -999.0; // user input

   // parameter for assymetry
   double tau = 1.47; 
 
   int i; // program interface
   for (i = 1; i < argc; i++) {
      string arg = argv[i];

      int n = 1;
      while (i+n < argc && argv[i+n][0] != '-') n++;

      if (arg == "-pqr") {
         if (n == 2) pqrFile = argv[++i];
         else cerr << "-pqr requires a file parameter" << endl;
      }
      else if (arg == "-neck") {
         if (n==2) neckFile = argv[++i];
         else cerr << "-neck requires file parameter" << endl;
      }
      else if (arg == "-neckRw") {
         if (n == 2) Neck_Rw = atof(argv[++i]);
         else cerr << "-neckRw needs a double value" << endl;
      }
      else if (arg == "-psf") {
         if (n == 2) psfFile = argv[++i];
         else cerr << "-psf requires file parameter" << endl;
      }
      else if (arg == "-prmtop") {
         if (n == 2) prmtopFile = argv[++i];
         else cerr << "-prmtop needs a file parameter." << endl;
      }
      else if (arg == "-chunk") {
         if (n == 3) {
            chunkFile = argv[++i];
            chunkDepth = atoi(argv[++i]);
         }
         else if (n == 2) {
            chunkFile = argv[++i];
            chunkDepth = 3;
            i++;
         } 
         else cerr << "-chunk requires chunk file and depth parameters" << endl;
      }
      else if (arg == "-AR6") {
         analytic = true;
         if (n != 1) cerr<<"-AR6, additional parameters ignored"<<endl;
      }
      else if (arg == "-NSR6") {
         molecularIntegrationR6 = true;
         if (n != 1)cerr<<"-NSR6, additional parameters ignored"<<endl;
      }
      else if (arg == "-A") {
         if (n == 2) {
            A = atof(argv[++i]);
            elsize_input = true;
         }
         else cerr << "-A needs a double value." << endl;
      }
      else if (arg == "-alpha") {
         if (n == 2) alpha = atof(argv[++i]);
         else cerr << "-alpha needs a double value." << endl;
      }
      else if (arg == "-tau") {
         if (n == 2) tau = atof(argv[++i]);
         else cerr << "-tau needs a double value." << endl;
      } 
      else if (arg == "-beta1") {
         if (n == 2) beta = atof(argv[++i]);
         else cerr << "-beta needs a double value." << endl;
      }
      else if (arg == "-beta2") {
         if (n == 2) gamma = atof(argv[++i]);
         else cerr << "-gamma needs a double value." << endl;
      }
      else if (arg == "-ROH") {
         if (n==2) ROH = atof(argv[++i]);
         else cerr << "-ROH needs a double value." << endl;
      }
      else if (arg == "-Svdw") {
         if (n == 2) delta = atof(argv[++i]);
         else cerr << "-delta needs a double value." << endl;
      }
      else if (arg == "-Sneck") {
         if (n == 2) omega = atof(argv[++i]);
         else cerr << "-omega needs a double value." << endl;
      }
      else if (arg == "-B") {
         if (n == 2) {  
            B_in = atof(argv[++i]);
            B = B_in;
         }
         else cerr << "-B needs a double value." << endl;
      }
      else if (arg == "-surf_density") {
         if (n == 2) density = atoi(argv[++i]);
         else cerr << "-surf_density needs a double value." << endl;
      }
      else if (arg == "-probe") {
         if (n == 2) {  ProbeRadius = atof(argv[++i]);
                        ProbeRadius_in  = true  ; }
         else cerr << "-probe needs a file parameter  " << endl;
      }
      else if (arg == "-epsin") {
         if (n == 2) diel_int = atof(argv[++i]);
         else cerr << "-epsin needs a double value" << endl;
      }
      else if (arg == "-epsout") {
         if (n == 2) diel_ext = atof(argv[++i]);
         else cerr << "-epsout needs a double value" << endl;
      }
      else if (arg == "-kappa") {
         if (n == 2) kappa = atof(argv[++i]);
         else cerr << "-kappa needs a double value" << endl;
      }
      else if (arg == "-gamma") {
         if (n == 2) { 
            gammaSASA_in = atof(argv[++i]);
            gammaSASA = gammaSASA_in ;
         }            
         else cerr << "-gamma needs a double value" << endl;
      }
      else if (arg == "-radii") {
         if (n == 2) rinvFile = argv[++i];
         else cerr << "-radii requires file parameter" << endl;
      }
      else if (arg == "-rinv_in") {
         if (n == 2) rinvIn = argv[++i];
         else cerr << "-rinv_in requires file parameter" << endl;
      }
      else if (arg == "-energy") {
         if (n == 2) etype = argv[++i];
         else cerr << "-out requires file parameter" << endl;
      }
      else if (arg == "-nolabel") {
         nolabel = true;
         if (n !=1) cerr<<"-nolabel additional parameters ignored"<<endl;
      }
      else if (arg == "-decom_tot") {
         if (n == 2) {
            solv_only = false;
            DGijFile = argv[++i];
         }
         else cerr << "-decom_tot needs a file parameter." << endl;
      }
      else if (arg == "-decom_solv") {
         if (n == 2) {
            solv_only = true;
            DGijFile = argv[++i];
         }
         else cerr << "-decom_solv needs a file parameter." << endl;
      }
      else if (arg == "-inradii" ) {
         inradii = argv[++i] ; 
         if (n !=2) cerr<<"-inradii, additional parameters ignored"<<endl;
      }
      else if (arg == "-chagb" ) {
         chagb_flag = true; 
         if (n !=1) cerr<<"-chagb, additional parameters ignored"<<endl;
      }
      else if (arg == "-canonical" ) {
         gb_flag = true; 
         if (n !=1) cerr<<"-canonical, additional parameters ignored"<<endl;
      }
      else if (arg == "-vdwprm" ) {
         ParameterFile_in = argv[++i] ;
         ParameterFile =  ParameterFile_in ;
         if (n !=2) cerr<<"-vdwprm, additional parameters ignored"<<endl;
      }
      else  //Invalid parameter
         cerr << "Invalid parameter " << i << ": " << arg << endl;
   }

   Molecule m;
   neckParams.WaterRadius(Neck_Rw);
   
   // Load PQR file, exit if not available
   if (pqrFile != "") {
      m.loadPQRFile(pqrFile, XYZformat, chagb_flag );
   } 
   else {
      cerr << "pqr file is required  ...exiting"<<endl;
      exit(1);
   }

   if ( (m.size() < 50) && (B  > 0) )  {
      cerr << " WARNING: B=0 is recommended for molecules with less than 50 atoms" << endl;
   }

   //  Load topology files 
   if (prmtopFile != "") {
      m.loadPRMTOPFile(prmtopFile);
   }
   else if (psfFile != "") {
      m.loadPSFFile(psfFile);
   }

   // Select the radii set and  parameter (gammaSASA + cdW)
   if ( inradii == "zap9" ) { // zap9 radii
      if (chagb_flag ) {
         cerr << "chagb and zap9 is not recommended ...exiting"<<endl;
         exit(1);
      }

      if (prmtopFile != "") {
         m.ZAP9radii(prmtopFile);
      } else {
         cerr<<"-zap9, the topology file is needed ..exiting"<<endl;
         exit(1);
      }

      if (gammaSASA_in == -999.0 )  gammaSASA = 0.01; 
      if (ParameterFile_in == "" )  ParameterFile = "LJ_zap9.dat";
      if ( B_in == -1.0 )  B = 0;
   }
   else if (inradii == "bestgb" ) { // optimum found for gb with 10 parmeterts
      if (chagb_flag ) {
         cerr << "bestgb and chagb  is not recommended ...exiting"<<endl;
         exit(1);
      }
      m.BestGBradii(chagb_flag);

      if (gammaSASA_in == -999.0 )  gammaSASA = 0.0104;
      if (ParameterFile_in == "" )  ParameterFile = "LJ_bestgb.dat";
      //if ( B_in == -1.0 )  B = 0;      
   } 
           
   else if( chagb_flag  ) {

      if (inradii != "inpqr" )
          m.BestGBradii(chagb_flag);  

      if (gammaSASA_in == -999.0 )  gammaSASA = 0.0178;
      if (ParameterFile_in == "" )  ParameterFile = "LJ_chagb.dat";
      //if ( B_in == -1.0 )  B = 0;
      if ( !ProbeRadius_in ) {ProbeRadius = 0.88 ;}
      else
      {ProbeRadius =ProbeRadius - 0.52;}
   }

   // If electrostatic size,A, is not provided by user , compute..
   if ( ! elsize_input ) A = elsize::ComputeElSize(pqrFile); 

   // compute Effective R6 Born Radii
   if (rinvIn == "") {
      if ( !analytic && ! molecularIntegrationR6 ) {
         cerr << "-NSR6 or -AR6 must be specified ...exiting" << endl;
	 exit(1);
      }
	
      // find neighboring atoms from topology files
      if (pqrFile != "" && analytic) {		
         if ( prmtopFile == "" && psfFile == "" ) {
            cerr << "Warning: chunk configuration will be"; 
            cerr << " set up from the pqr file, we strongly ";
            cerr << "recommend to use topology files for ";
            cerr << "higher accuracy" << endl;
	 }					
	 m.initalizeNeighbors();
      }

      // Read neck parameters
      if (neckFile != "")
         neckParams.loadFromFile(neckFile);
      else 
         neckParams.loadFromFile("temp.neck");
  		
      // Compute chunks
      if (pqrFile != "" && analytic ) {
         string s = prmtopFile;
	 if (s == "") s = pqrFile;
	 m.initalizeChunks(chunkFile, s, chunkDepth);
      }

      // compute radii
      if (pqrFile != "" && analytic ) // use AR6
         Analytic::calculate(m, neckParams, A, B, alpha, beta, gamma, delta, omega);
      else if ( pqrFile != "" && molecularIntegrationR6) // use NSR6
	 MolecularIntegration::EntireMoleculeR6(m, density, ProbeRadius, B);
      
   }        
   else  // Read effective radii from file
      m.loadRinv(rinvIn);
	
   // print the effective radii 
   if (rinvFile != "") 
      dispMolecularIntegrationR6(m, rinvFile);

   // Once the R6 Born radii is computed, calculted energies and display
   double polar_solv, nonpolar_solv, total_solv, coulomb, cavity_solv, vdw_solv;
   if (etype == "off" )
      return 0;
   
   // LABELs dor displaying
   string pol_label="", nonpol_label="", totsol_label="",
          coul_label="", total_label="" ;

   if ( !nolabel) {
      pol_label="POLAR_SOLV ";
      nonpol_label="NONPOLAR_SOLV ";
      totsol_label="TOTAL_SOLV ";
      coul_label="COULOMB ";
      total_label = "TOTAL_EELEC " ;
      cout <<"Energies(kcal/mol): ";
   }
        
   // displaying polar solvation enrgy only
   if (etype == "polar") {
    if (gb_flag) {
     if (chagb_flag){
        polar_solv = GBorn::chagbcan(m,kappa,diel_ext,diel_int,tau,ROH);}
          else{polar_solv = GBorn::gb(m, kappa , diel_ext, diel_int);}
      }
      else{  
      if ( chagb_flag) {polar_solv = GBorn::chagb(m, A, kappa , diel_ext, diel_int, tau, ROH);}
      else
      {polar_solv = GBorn::alpb(m, A, kappa , diel_ext, diel_int);}
      }
      cout << pol_label << polar_solv << endl;
   }
   // displaying non polar only
   else if (etype == "nonpolar") {

      if ( ParameterFile != "" )
         vdw_solv = GBorn::VDWenergy( m, ProbeRadius, ParameterFile) ;

      if (analytic) 
         cavity_solv = gammaSASA*(GBorn::SASA(m,ProbeRadius)) ; 
      else if ( molecularIntegrationR6  ) 
         cavity_solv = gammaSASA*(GBorn::GetSASFromFile("msms.out"));
	 
      cout << nonpol_label <<  cavity_solv + vdw_solv << endl;
   }
        
   // displaying total solvation energy (first column)
   //            polar solvation (second column)
   //            nonpolar solvation (third column
   else if (etype=="total_solv") {
    if (gb_flag) {
     if (chagb_flag){
        polar_solv = GBorn::chagbcan(m,kappa,diel_ext,diel_int,tau,ROH);}
          else{polar_solv = GBorn::gb(m, kappa , diel_ext, diel_int);}
      }
      else{  
      if ( chagb_flag) {polar_solv = GBorn::chagb(m, A, kappa , diel_ext, diel_int, tau, ROH);}
      else
      {polar_solv = GBorn::alpb(m, A, kappa , diel_ext, diel_int);}
      }
         if ( ParameterFile != "" )
         vdw_solv = GBorn::VDWenergy( m, ProbeRadius, ParameterFile) ;           
	    
      if (analytic ) 
          cavity_solv = gammaSASA* (GBorn::SASA(m, ProbeRadius)) ;
      else if ( molecularIntegrationR6  ) 
          cavity_solv = gammaSASA* (GBorn::GetSASFromFile("msms.out"));

      cout << totsol_label << polar_solv + cavity_solv + vdw_solv << "  ";
            cout << pol_label << polar_solv << "  ";
            cout << nonpol_label << cavity_solv + vdw_solv << endl;
   }
   //  Coulumb electrostatic energy
   else if (etype=="coulomb") { 
      coulomb = GBorn::Coulomb(m,diel_int) ;
      cout << coul_label << coulomb << endl;
   }
   // compute all energies
   else {
      if  ( !(etype == "all")  &&  !(etype == "") ) {
          cerr << "WARNING: energy type (-etype) not recongnized,"
               << "  computing all energies ..." << endl;
      }
    if (gb_flag) {
     if (chagb_flag){
        polar_solv = GBorn::chagbcan(m,kappa,diel_ext,diel_int,tau,ROH);}
          else{polar_solv = GBorn::gb(m, kappa , diel_ext, diel_int);}
      }
      else{  
      if ( chagb_flag) {polar_solv = GBorn::chagb(m, A, kappa , diel_ext, diel_int, tau, ROH);}
      else
      {polar_solv = GBorn::alpb(m, A, kappa , diel_ext, diel_int);}
      }

      if ( ParameterFile != "" ) 
         vdw_solv = GBorn::VDWenergy( m, ProbeRadius, ParameterFile) ;
 
      if (analytic ) 
         cavity_solv = gammaSASA* (GBorn::SASA(m, ProbeRadius)) ;
      else if ( molecularIntegrationR6  ) 
         cavity_solv = gammaSASA* (GBorn::GetSASFromFile("msms.out"));
                       
      nonpolar_solv  =  cavity_solv + vdw_solv;
      coulomb = GBorn::Coulomb(m,diel_int) ;

      cout << total_label << polar_solv + nonpolar_solv + coulomb << "  ";
      cout << totsol_label  <<polar_solv + nonpolar_solv << "  ";
      cout << pol_label << polar_solv << "  ";
      cout << nonpol_label << nonpolar_solv << "  ";
      cout << coul_label << coulomb << endl;        
   }

   if ( DGijFile != "") {
 if (gb_flag) {
     if (chagb_flag){
        polar_solv = GBorn::chagbcan_ij(m,kappa,diel_ext,diel_int,tau,ROH,DGijFile,solv_only);}
          else{GBorn::gb_ij(m,kappa,diel_ext,diel_int, DGijFile, solv_only );}
      }
      else{  
      if ( chagb_flag) {
         GBorn::chagb_ij(m,A,kappa,diel_ext,diel_int,tau,ROH, DGijFile, solv_only );
      }
      else
      {GBorn::alpb_ij(m, A, kappa , diel_ext, diel_int, DGijFile, alpha_alpb, solv_only ) ;}
      }

   }
   return 0;
}

// Display the calculated R6^(-1) born Radii
bool dispMolecularIntegrationR6(Molecule &m, string rinvFile) {
   ofstream out(rinvFile.c_str());
   if (!out) {
      cerr << "problems opening rinvFile" << endl ;
      return false;
   }

   Atoms::iterator i;
   for (i = m.begin(); i != m.end(); i++) {
      out << i->Number() << "\t" << i->MolecularR6Sum() << endl;
   }
   return out;
}
