double VanDerWaalsVolume(const Molecule &m);
bool noIntersections(Molecule &m, int num);
void LoadVolumeData(Molecule &m, string fileName);
void modificationsByType(Molecule &m);
Atoms involvedInItersection(Molecule &m, int num, int atomNum);
void maxIntersections(Molecule &m);

//-----------------------------------------------------------------------------
void Foo(Molecule &m, NeckParams &neckParams, double alpha, double beta, double gamma, double delta,
 double omega) {
  double R_w = 1.4;
  double startRadiusI, endRadiusI, startRadiusJ, endRadiusJ, stepRadiusI,
   stepRadiusJ, startDistance, endDistance, stepDistance;
  startRadiusI = endRadiusI = 1.2;
  startRadiusJ = endRadiusJ = 1.2;
  stepRadiusI = stepRadiusJ = 0.1;
  startDistance = 0.01;
  stepDistance = 0.01;
  //endDistance = 0.05;

  double i, j, d;

  for (i = startRadiusI; i <= endRadiusI; i += stepRadiusI) {
    for (j = startRadiusJ; j <= endRadiusJ; j += stepRadiusJ) {
      endDistance = 1+i+j+2*R_w;
      for (d = startDistance; d <= endDistance; d += stepDistance) {
        Atom a_i(0,0,0,i,1);
        Atom a_j(d,0,0,j,2);
        Molecule m; m.addAtom(a_i); m.addAtom(a_j);
        MolecularIntegration::EntireMoleculeR6(m, 40, 1.4);
        Atoms::iterator itr = m.begin();
        double R = itr->MolecularR6Sum();
        double integral = (4.*M_PI)/3. * (1/(i*i*i) - R*R*R);
        double term = j / (d*d - j*j);
        double gry;
        if (d > i+j) {
          gry = ( (4.*M_PI/3.) ) * term*term*term;
        }
        else {
          double v0 = 16 * d;
          double v1 = d + 3*j;
          double v2 = d + j;
          double v3 = 3*(j*j - i*i - (d-i)*(d-i))+2*d*i;
          double v4 = i*i*i*i;
          gry = ( (4.*M_PI)/3. ) * (1/v0) * (v1/(v2*v2*v2) + v3/v4);
        }

        double Aij = 9.47808e-05;
	double neck = Aij*pow(d,4)*pow(i+j+2*1.4-d,4);
        cout << d << "\t" << integral - gry << "\t" << neck << endl;
      }
    }
  }
}

void Bar(Molecule &m) {
	MolecularIntegration::EntireMoleculeR6(m, 40, 1.4);
	Atoms::iterator i;
	for (i = m.begin(); i != m.end(); i++) {
		cout << i->Number() << "\t" << i->MolecularR6Sum() << endl;
	}
}

bool noIntersections(Molecule &m, int num) {
	cout << "*****Testing overlaps*****" << endl;
	if (num > m.size()) {
		cout << "Invalid param num: " << num << endl;
		return true;
	}
	bool result = true;

	Atoms test;
	Atoms::iterator i, j;
	int pos, count = 1;
	for (pos = 0; pos < num-1; pos++) {
		test.push_back(*(m.begin() + pos));
	}
	test.push_back(*(m.begin()+pos));

	while (true) {
		int n;
		i = test.end()-1;
		while (i != test.end()) {
			if (i != test.begin()) {
				for (n = i->Number(); n < m.size(); n++) {
					for (j = test.begin(); j != i; j++) {
						if ((m.begin()+n)->distance(*j) > (m.begin()+n)->Radius() + j->Radius()) break;
					}
					if (j == i) break;
				}
			}
			else {
				n = i->Number();
			}
			if (n < m.size()) {
				i = replace(test, i, *(m.begin()+n));
				i++;
				if ((i-1)->Number() >= m.size()) break;
				if (i != test.end()) i = replace(test, i, *(m.begin()+(i-1)->Number()));
			}
			else {
				if (i == test.begin()) break;
				i--;
			}
		}	
		if (i != test.end()) break;

		cout << count << ". ";
		count++;
		for (i = test.begin(); i != test.end(); i++) {
			if (i+1 == test.end()) cout << i->Number() << endl;
			else cout << i->Number() << ", ";
		}
		
	}
	return result;
}

double VanDerWaalsVolume(const Molecule &m) {
	double sum = 0;
	Atoms::const_iterator i;
	for (i = m.begin(); i != m.end(); i++) {
		double R = i->Radius();
		sum += 1.33333 * M_PI * R*R*R;
	}
	return sum;
}
